from dataclasses import fields
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import streamlit as st

from helix.options.choices.ui import DATA_SPLITS, PLOT_FONT_FAMILIES
from helix.options.data import DataSplitOptions
from helix.options.enums import DataSplitMethods, ExecutionStateKeys, PlotOptionKeys
from helix.options.file_paths import (
    data_options_path,
    data_preprocessing_options_path,
    execution_options_path,
    fi_options_path,
    ml_options_path,
    plot_options_path,
)
from helix.options.plotting import PlottingOptions
from helix.services.configuration import (
    load_data_options,
    load_data_preprocessing_options,
    load_execution_options,
    load_fi_options,
    load_ml_options,
    load_plot_options,
)


@st.experimental_fragment
def plot_options_box(plot_opts: PlottingOptions | None = None):
    """Expander containing the options for making plots

    Args:
        plot_opts (PlottingOptions, optional): Plot options loaded from configuration.
            If provided, use these values instead of defaults.
    """
    with st.expander("Plot options", expanded=False):
        # Use configuration values if available, otherwise use defaults
        save = st.checkbox(
            "Save all plots",
            key=PlotOptionKeys.SavePlots,
            value=plot_opts.save_plots if plot_opts else True,
        )
        width = st.number_input(
            "Width",
            min_value=10,
            max_value=100,
            key=PlotOptionKeys.Width,
            value=plot_opts.width if plot_opts else 10,
        )
        height = st.number_input(
            "Height",
            min_value=10,
            max_value=100,
            key=PlotOptionKeys.Height,
            value=plot_opts.height if plot_opts else 10,
        )
        dpi = st.slider(
            "Image Resolution (DPI)",
            min_value=330,
            max_value=2000,
            value=plot_opts.dpi if plot_opts else 330,
            key=PlotOptionKeys.DPI,
        )
        rotate_x = st.number_input(
            "Angle to rotate X-axis labels",
            min_value=0,
            max_value=90,
            value=plot_opts.angle_rotate_xaxis_labels if plot_opts else 45,
            key=PlotOptionKeys.RotateXAxisLabels,
            disabled=not save,
        )
        rotate_y = st.number_input(
            "Angle to rotate Y-axis labels",
            min_value=0,
            max_value=90,
            value=plot_opts.angle_rotate_yaxis_labels if plot_opts else 0,
            key=PlotOptionKeys.RotateYAxisLabels,
            disabled=not save,
        )
        tfs = st.number_input(
            "Title font size",
            value=plot_opts.plot_title_font_size if plot_opts else 16,
            min_value=8,
            key=PlotOptionKeys.TitleFontSize,
            disabled=not save,
        )
        afs = st.number_input(
            "Axis font size",
            value=plot_opts.plot_axis_font_size if plot_opts else 12,
            min_value=8,
            key=PlotOptionKeys.AxisFontSize,
            disabled=not save,
        )
        ats = st.number_input(
            "Axis tick size",
            value=plot_opts.plot_axis_tick_size if plot_opts else 10,
            min_value=8,
            key=PlotOptionKeys.AxisTickSize,
            disabled=not save,
        )
        # Get valid style name, defaulting to seaborn-v0_8-whitegrid
        default_style = (
            "seaborn-v0_8-whitegrid"
            if "seaborn-v0_8-whitegrid" in plt.style.available
            else plt.style.available[0]
        )
        style_index = (
            plt.style.available.index(plot_opts.plot_colour_scheme)
            if plot_opts and plot_opts.plot_colour_scheme in plt.style.available
            else plt.style.available.index(default_style)
        )
        cs = st.selectbox(
            "Colour scheme",
            options=plt.style.available,
            key=PlotOptionKeys.ColourScheme,
            disabled=not save,
            index=style_index,
        )
        colormap_index = (
            plt.colormaps().index(plot_opts.plot_colour_map)
            if plot_opts
            else plt.colormaps().index("viridis")
        )
        cm = st.selectbox(
            "Colour map",
            options=plt.colormaps(),
            key=PlotOptionKeys.ColourMap,
            index=colormap_index,
            disabled=not save,
        )
        font_index = (
            PLOT_FONT_FAMILIES.index(plot_opts.plot_font_family)
            if plot_opts
            else PLOT_FONT_FAMILIES.index(PLOT_FONT_FAMILIES[1])  # sans-serif
        )
        font = st.selectbox(
            "Font",
            options=PLOT_FONT_FAMILIES,
            key=PlotOptionKeys.FontFamily,
            disabled=not save,
            index=font_index,
        )
        if save:
            """Here we show a preview of plots with the selected colour style
            colour map, font size and style, etc"""

            plt.rcParams["image.cmap"] = cm  # default colour map
            plt.rcParams["axes.titlesize"] = tfs
            plt.rcParams["axes.labelsize"] = afs
            plt.rcParams["font.family"] = font
            plt.rcParams["xtick.labelsize"] = ats
            plt.rcParams["ytick.labelsize"] = ats

            st.write("### Preview of the selected styles")
            plt.style.use(cs)
            # Generate some random data for demonstration
            arr = np.random.normal(1, 0.5, size=100)
            # Create a violin plot
            data = pd.DataFrame({"A": arr, "B": arr, "C": arr})
            fig, ax = plt.subplots(figsize=(width, height), dpi=dpi)
            sns.violinplot(data=data, ax=ax)
            ax.set_title("Title")
            ax.set_xlabel("X axis")
            ax.set_ylabel("Y axis")
            ax.tick_params(labelsize=ats)
            ax.set_xticklabels(ax.get_xticklabels(), rotation=rotate_x)
            ax.set_yticklabels(ax.get_yticklabels(), rotation=rotate_y)
            st.pyplot(fig, clear_figure=True)
            fig.clear()
            # Create a figure and axis (object-oriented approach)
            fig_cmap = plt.figure(dpi=dpi)
            ax_cmap = fig_cmap.add_subplot(111)

            # Create a scatter plot to show how the colour map is applied
            scatter_plot = ax_cmap.scatter(arr, arr / 2, c=arr)
            fig_cmap.colorbar(scatter_plot, ax=ax_cmap, label="Mapped Values")
            ax_cmap.set_title("Colour Map Preview")
            # Display the figure
            st.pyplot(fig_cmap, clear_figure=True)
            fig.clear()


@st.experimental_fragment
def data_split_options_box(manual: bool = False) -> DataSplitOptions:
    """Component for configuring data split options.

    TODO: in a future PR remove the `manual` param when we can
    perform holdout and kfold with grid search.

    Args:
        manual (bool): Using manual hyperparameter setting?

    Returns:
        DataSplitOptions: The options used to split the data.
    """

    st.subheader("Configure data split method")
    if manual:
        data_split = st.selectbox("Data split method", DATA_SPLITS)
    else:
        data_split = DataSplitMethods.NoSplit
    n_bootsraps = None
    k = None
    if data_split.lower() == DataSplitMethods.Holdout:
        n_bootsraps = st.number_input(
            "Number of bootstraps",
            min_value=1,
            value=10,
            key=ExecutionStateKeys.NumberOfBootstraps,
        )
    else:
        k = st.number_input(
            "Number of folds in Cross-Validation k",
            min_value=1,
            value=5,
            help="k is the number of folds in Cross-Validation",
        )
    test_size = st.number_input(
        "Test size (fraction of data)",
        min_value=0.0,
        max_value=1.0,
        value=0.2,
        help="The fraction (between 0 and 1) to reserve for testing models on unseen data.",
    )

    return DataSplitOptions(
        method=data_split, n_bootstraps=n_bootsraps, k_folds=k, test_size=test_size
    )


def display_options(experiment_path: Path) -> None:
    """Display the options in the sidebar."""

    # Load all options
    options_dict = {
        "Execution Options": load_execution_options(
            execution_options_path(experiment_path)
        ),
        "Data Options": load_data_options(data_options_path(experiment_path)),
        "Plotting Options": load_plot_options(plot_options_path(experiment_path)),
        "Preprocessing Options": load_data_preprocessing_options(
            data_preprocessing_options_path(experiment_path)
        ),
        "Machine Learning Options": load_ml_options(ml_options_path(experiment_path)),
        "Feature Importance Options": load_fi_options(fi_options_path(experiment_path)),
        # "Fuzzy Options": load_fuzzy_options(fuzzy_options_path(experiment_path)),
    }

    with st.expander("Show Experiment Options", expanded=False):

        # Display Execution Options (dataclass format)
        for option in options_dict:

            if options_dict[option] is None:
                continue

            st.write(f"### {option}")
            execution_options = options_dict[option]

            data = {
                "Variable": [field.name for field in fields(execution_options)],
                "Value": [
                    getattr(execution_options, field.name)
                    for field in fields(execution_options)
                ],
            }

            st.table(pd.DataFrame(data).set_index("Variable"))



import pandas as pd
import streamlit as st

from helix.options.choices.ui import NORMALISATIONS, TRANSFORMATIONS_Y
from helix.options.enums import (
    DataPreprocessingStateKeys,
    ProblemTypes,
    TransformationsY,
)


@st.experimental_fragment
def preprocessing_opts_form(data: pd.DataFrame, problem_type: ProblemTypes):
    st.write("## Data preprocessing options")

    st.write("### Data normalisation")

    st.write(
        """
        If you select **"Standardisation"**, your data will be normalised by subtracting the
        mean and dividing by the standard deviation for each feature. The resulting transformation has a
        mean of 0 and values are between -1 and 1.

        If you select **"Minmax"**, your data will be scaled based on the minimum and maximum
        value of each feature. The resulting transformation will have values between 0 and 1.

        If you select **"Mean centering"**, your data will be transformed by subtracting the mean
        of each feature, so that the resulting distribution has a mean of 0 but is not scaled by variance.

        If you select **"Mean centering and Poisson scaling"**, your data will first be mean-centered,
        then scaled by dividing each feature by the square root of its mean. This is useful when the
        variance of your data is approximately proportional to the mean, as in Poisson-like distributions.

        If you select **"Pareto scaling"**, your data will be scaled by dividing each feature by
        the square root of its standard deviation, without mean-centering. This reduces the relative
        importance of high-variance features while keeping some of their weight.

        If you select **"None"**, the data will not be normalised.
        """
    )

    st.write("#### Normalisation method for independent variables")

    st.selectbox(
        "Normalisation",
        NORMALISATIONS,
        key=DataPreprocessingStateKeys.IndependentNormalisation,
        index=len(NORMALISATIONS) - 1,  # default to no normalisation
    )

    st.write("#### Transformation method for dependent variable")

    st.write(
        """
        If you select **Log**, the dependent variable will transformed with the natural logarithm (`ln` / `log e`). If the minimum of your dependent variable is less than or equal to 0, it is first transformed by `(y - min(y)) + 1` to make all values positive, followed by the natural logarithm.

        If you select **Square-root**, the dependent variable is transformed by taking the sqare root of each value. If the minimum of your dependent variable is less than 0, it is first transformed by `y - min(y)` to make all values at least 0, followed by the square root.

        If you select **"Minmax"**, your dependent variable will be scaled based on the minimum and maximum value of each feature. The resulting transformation will have values between 0 and 1.

        If you select **"Standardisation"**, your dependent variable will be normalised by subtracting the mean and dividing by the standard deviation for each feature. The resulting transformation has a mean of 0 and values are between -1 and 1.

        If you select **"None"**, the data will not be normalised.
        """
    )

    # disable dependent variable transformation for classifications
    no_transformation_y = problem_type.lower() == ProblemTypes.Classification
    if no_transformation_y:
        st.warning(
            "Transformation of dependent variables is disabled for **classification** experiments.",
            icon="⚠️",
        )
    transformation_y = st.selectbox(
        "Transformations",
        TRANSFORMATIONS_Y,
        key=DataPreprocessingStateKeys.DependentNormalisation,
        index=len(TRANSFORMATIONS_Y) - 1,  # default to no transformation,
        disabled=no_transformation_y,
    )

    if (
        transformation_y.lower() == TransformationsY.Log
        or transformation_y.lower() == TransformationsY.Sqrt
    ):
        if (
            data.iloc[:, -1].min() <= 0
        ):  # deal with user attempting this transformations on negative values
            st.warning(
                "The dependent variable contains negative values. Log and square root transformations require positive values."
            )
            if st.checkbox(
                "Proceed with transformation. This option will add a constant to the dependent variable to make it positive.",
                key=DataPreprocessingStateKeys.ProceedTransformation,
            ):
                pass
            else:
                st.stop()

    st.write("### Feature selection")

    st.write("#### Check the feature selection algorithms to use")

    variance_disabled = True
    if st.checkbox(
        "Variance threshold",
        key=DataPreprocessingStateKeys.VarianceThreshold,
        help="Delete features with variance below a certain threshold",
    ):
        variance_disabled = False
    st.number_input(
        "Set threshold",
        min_value=0.0,
        max_value=1.0,
        value=0.1,
        key=DataPreprocessingStateKeys.ThresholdVariance,
        disabled=variance_disabled,
    )

    correlation_disabled = True
    if st.checkbox(
        "Correlation threshold",
        key=DataPreprocessingStateKeys.CorrelationThreshold,
        help="Delete features with correlation above a certain threshold",
    ):
        correlation_disabled = False
    st.number_input(
        "Set threshold",
        min_value=0.0,
        max_value=1.0,
        value=0.8,
        key=DataPreprocessingStateKeys.ThresholdCorrelation,
        disabled=correlation_disabled,
    )

    lasso_disabled = True
    if st.checkbox(
        "Lasso Feature Selection",
        key=DataPreprocessingStateKeys.LassoFeatureSelection,
        help="Select features using Lasso regression",
    ):
        lasso_disabled = False
    st.number_input(
        "Set regularisation term",
        min_value=0.0,
        value=0.05,
        key=DataPreprocessingStateKeys.RegularisationTerm,
        disabled=lasso_disabled,
    )



from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
import seaborn as sns
import shap
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.ticker import FormatStrFormatter
from sklearn.manifold import TSNE
from sklearn.metrics import ConfusionMatrixDisplay, RocCurveDisplay

from helix.options.plotting import PlottingOptions


def plot_target_variable_distribution(
    data: pd.DataFrame,
    show_kde: bool,
    n_bins: int,
    plot_opts: PlottingOptions,
    dep_var_name: str,
) -> Figure:
    """
    Create a distribution plot for the target variable.

    Args:
        data (pd.DataFrame): The DataFrame containing the target variable.
        show_kde (bool): Whether to show the kernel density estimate.
        n_bins (int): Number of bins for the histogram.
        plot_opts (PlottingOptions): The plotting options.
        dep_var_name (str): The name of the dependent variable.

    Returns:
        Figure: The distribution plot figure.

    """

    plt.style.use(plot_opts.plot_colour_scheme)
    plt.figure(figsize=(plot_opts.width, plot_opts.height), dpi=plot_opts.dpi)

    displot = sns.displot(
        data=data,
        x=data.columns[-1],
        kde=show_kde,
        bins=n_bins,
        height=plot_opts.height,
        aspect=plot_opts.width / plot_opts.height,
        color=plot_opts.plot_colour,
    )

    title = (
        plot_opts.plot_title if plot_opts.plot_title else f"{dep_var_name} Distribution"
    )

    plt.title(
        title,
        fontdict={
            "family": plot_opts.plot_font_family,
            "fontsize": plot_opts.plot_title_font_size,
        },
    )

    x_label = plot_opts.xaxis_label if plot_opts.xaxis_label else dep_var_name

    plt.xlabel(
        x_label,
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )

    y_label = plot_opts.yaxis_label if plot_opts.yaxis_label else "Frequency"

    plt.ylabel(
        y_label,
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )

    plt.xticks(
        rotation=plot_opts.angle_rotate_xaxis_labels,
        fontsize=plot_opts.plot_axis_tick_size,
        family=plot_opts.plot_font_family,
    )
    plt.yticks(
        rotation=plot_opts.angle_rotate_yaxis_labels,
        fontsize=plot_opts.plot_axis_tick_size,
        family=plot_opts.plot_font_family,
    )

    return displot


def plot_correlation_heatmap(corr: pd.DataFrame, plot_opts: PlottingOptions) -> Figure:
    """
    Create a correlation heatmap for the given correlations DataFrame.

    Args:
        corr_data (pd.DataFrame): The DataFrame containing the correlation data to plot.
        plot_opts (PlottingOptions): The plotting options.

    Returns:
        Figure: The correlation heatmap figure.
    """

    # Only mask if rows == cols in the same order (true symmetric matrix view)
    is_square = corr.shape[0] == corr.shape[1]
    same_labels_same_order = is_square and corr.index.equals(corr.columns)

    mask = np.triu(np.ones_like(corr, dtype=bool)) if same_labels_same_order else None

    plt.style.use(plot_opts.plot_colour_scheme)
    fig, ax = plt.subplots(
        figsize=(plot_opts.width, plot_opts.height),
        dpi=plot_opts.dpi,
    )

    sns.heatmap(
        corr,
        mask=mask,
        cmap=plot_opts.plot_colour_map,
        vmax=1.0,
        vmin=-1.0,
        center=0,
        square=True,
        linewidths=0.5,
        annot=True,
        fmt=".2f",
        cbar_kws={
            "shrink": 0.5,
            "label": "Correlation Coefficient",
            "format": "%.1f",
            "aspect": 30,
            "drawedges": True,
        },
        annot_kws={
            "size": plot_opts.plot_axis_tick_size,
            "family": plot_opts.plot_font_family,
        },
        xticklabels=True,  # Ensure x-axis labels are shown
        yticklabels=True,  # Ensure y-axis labels are shown
        ax=ax,
    )

    title = plot_opts.plot_title if plot_opts.plot_title else "Correlation Heatmap"

    ax.set_title(
        title,
        fontsize=plot_opts.plot_title_font_size,
        family=plot_opts.plot_font_family,
        pad=20,
        wrap=True,
    )

    x_label = plot_opts.xaxis_label if plot_opts.xaxis_label else None

    y_label = plot_opts.yaxis_label if plot_opts.yaxis_label else None

    if x_label:
        ax.set_xlabel(
            x_label,
            fontsize=plot_opts.plot_axis_font_size,
            family=plot_opts.plot_font_family,
        )
    if y_label:
        ax.set_ylabel(
            y_label,
            fontsize=plot_opts.plot_axis_font_size,
            family=plot_opts.plot_font_family,
            rotation=plot_opts.angle_rotate_yaxis_labels,
        )

    plt.xticks(
        rotation=plot_opts.angle_rotate_xaxis_labels,
        ha="right",
        fontsize=plot_opts.plot_axis_tick_size,
        family=plot_opts.plot_font_family,
    )
    plt.yticks(
        rotation=plot_opts.angle_rotate_yaxis_labels,
        fontsize=plot_opts.plot_axis_tick_size,
        family=plot_opts.plot_font_family,
    )

    plt.tight_layout()

    return fig


def create_pairplot(
    pairplot_data: pd.DataFrame,
    plot_opts: PlottingOptions,
    exclude_corner: bool,
    kind: str,
    diag_kind: str,
) -> Figure:
    """
    Create a pairplot of the given DataFrame.

    Args:
        pairplot_data (pd.DataFrame): The DataFrame to plot.
        plot_opts (PlottingOptions): The plotting options.
        exclude_corner (bool): Whether to exclude the corner plot.
        kind (str): The type of plot to use for the pairplot.
        diag_kind (str): The type of plot to use on the diagonal.

    Returns:
        Figure: The pairplot figure.

    """

    plt.style.use(plot_opts.plot_colour_scheme)

    with plt.rc_context({"figure.dpi": plot_opts.dpi}):

        n_vars = len(pairplot_data.columns)
        aspect_ratio = plot_opts.width / plot_opts.height
        size_per_var = min(plot_opts.width, plot_opts.height) / n_vars

        pairplot = sns.pairplot(
            pairplot_data,
            height=size_per_var,
            aspect=aspect_ratio,
            corner=exclude_corner,
            kind=kind,
            diag_kind=diag_kind,
        )

        title = plot_opts.plot_title if plot_opts.plot_title else "Pairplot"

        pairplot.figure.suptitle(
            title,
            fontsize=plot_opts.plot_title_font_size,
            family=plot_opts.plot_font_family,
            y=1.02,  # Adjust title position to prevent overlap
        )

        for ax in pairplot.axes.flat:
            if ax is not None:
                # Rotate labels
                ax.set_xticklabels(
                    ax.get_xticklabels(),
                    rotation=plot_opts.angle_rotate_xaxis_labels,
                    family=plot_opts.plot_font_family,
                )
                ax.set_yticklabels(
                    ax.get_yticklabels(),
                    rotation=plot_opts.angle_rotate_yaxis_labels,
                    family=plot_opts.plot_font_family,
                )

                # Set tick font size
                ax.tick_params(labelsize=plot_opts.plot_axis_tick_size)

                # Set axis labels
                if ax.get_xlabel():
                    ax.set_xlabel(
                        ax.get_xlabel(),
                        fontsize=plot_opts.plot_axis_font_size,
                        family=plot_opts.plot_font_family,
                    )
                if ax.get_ylabel():
                    ax.set_ylabel(
                        ax.get_ylabel(),
                        fontsize=plot_opts.plot_axis_font_size,
                        family=plot_opts.plot_font_family,
                    )

        plt.tight_layout()

        return pairplot.figure


def create_tsne_plot(
    data: pd.DataFrame,
    normalised_data: pd.DataFrame,
    y: pd.Series | np.ndarray,
    plot_opts: PlottingOptions,
    random_state: int,
    perplexity: int,
) -> Figure:
    """
    Create a t-SNE plot for both normalised and original data.

    Args:
        data (pd.DataFrame): The original data to plot.
        normalised_data (pd.DataFrame): The normalised data to plot.
        y (pd.Series | np.ndarray): The target variable.
        plot_opts (PlottingOptions): The plotting options.
        random_state (int): Random state for reproducibility.
        perplexity (int): Perplexity parameter for t-SNE.

    Returns:
        Figure: The t-SNE plot figure containing two subplots.
    """

    tsne_normalised = TSNE(
        n_components=2, random_state=random_state, perplexity=perplexity
    )

    tsne_original = TSNE(
        n_components=2, random_state=random_state, perplexity=perplexity
    )

    X_embedded_normalised = tsne_normalised.fit_transform(normalised_data)
    X_embedded = tsne_original.fit_transform(data)

    df_normalised = pd.DataFrame(X_embedded_normalised, columns=["x", "y"])
    df_normalised["target"] = y

    df = pd.DataFrame(X_embedded, columns=["x", "y"])
    df["target"] = y

    plt.style.use(plot_opts.plot_colour_scheme)
    fig, ax = plt.subplots(
        1,
        2,
        figsize=(plot_opts.width, plot_opts.height),
        dpi=plot_opts.dpi,
    )

    sns.scatterplot(
        data=df_normalised,
        x="x",
        y="y",
        hue="target",
        palette=plot_opts.plot_colour_map,
        s=100,  # marker size
        alpha=0.6,  # transparency
        ax=ax[0],
    )

    # Customize first plot
    title = plot_opts.plot_title if plot_opts.plot_title else "t-SNE Plot"
    ax[0].set_title(
        title + " (Normalised Features)",
        fontsize=plot_opts.plot_title_font_size,
        family=plot_opts.plot_font_family,
        pad=20,  # Add padding above title
    )

    x_label = plot_opts.xaxis_label if plot_opts.xaxis_label else "t-SNE Component 1"
    ax[0].set_xlabel(
        x_label,
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )

    y_label = plot_opts.yaxis_label if plot_opts.yaxis_label else "t-SNE Component 2"
    ax[0].set_ylabel(
        y_label,
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )
    # Apply axis label rotations and styling for first plot
    ax[0].tick_params(
        axis="both", which="major", labelsize=plot_opts.plot_axis_tick_size
    )
    for label in ax[0].get_xticklabels():
        label.set_rotation(plot_opts.angle_rotate_xaxis_labels)
        label.set_family(plot_opts.plot_font_family)
    for label in ax[0].get_yticklabels():
        label.set_rotation(plot_opts.angle_rotate_yaxis_labels)
        label.set_family(plot_opts.plot_font_family)

    sns.scatterplot(
        data=df,
        x="x",
        y="y",
        hue="target",
        palette=plot_opts.plot_colour_map,
        s=100,  # marker size
        alpha=0.6,  # transparency
        ax=ax[1],
    )

    # Customize second plot
    ax[1].set_title(
        title + " (Original Features)",
        fontsize=plot_opts.plot_title_font_size,
        family=plot_opts.plot_font_family,
        pad=20,  # Add padding above title
    )
    ax[1].set_xlabel(
        x_label,
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )
    ax[1].set_ylabel(
        y_label,
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )
    # Apply axis label rotations and styling for second plot
    ax[1].tick_params(
        axis="both", which="major", labelsize=plot_opts.plot_axis_tick_size
    )
    for label in ax[1].get_xticklabels():
        label.set_rotation(plot_opts.angle_rotate_xaxis_labels)
        label.set_family(plot_opts.plot_font_family)
    for label in ax[1].get_yticklabels():
        label.set_rotation(plot_opts.angle_rotate_yaxis_labels)
        label.set_family(plot_opts.plot_font_family)

    # Adjust layout to prevent label cutoff
    plt.tight_layout()

    return fig


def plot_lime_importance(
    df: pd.DataFrame,
    plot_opts: PlottingOptions,
    num_features_to_plot: int,
    title: str,
) -> Figure:
    """Plot LIME importance.

    Args:
        df (pd.DataFrame): The LIME data to plot
        plot_opts (PlottingOptions): The plotting options.
        num_features_to_plot (int): The top number of features to plot.
        title (str): The title of the plot.

    Returns:
        Figure: The LIME plot.
    """
    # Calculate most important features
    most_importance_features = (
        df.abs()
        .mean()
        .sort_values(ascending=False)
        .head(num_features_to_plot)
        .index.to_list()
    )

    plt.style.use(plot_opts.plot_colour_scheme)
    fig, ax = plt.subplots(layout="constrained", dpi=plot_opts.dpi)

    sns.violinplot(
        data=df.loc[:, most_importance_features],
        fill=True,
        ax=ax,
        palette=plot_opts.plot_colour_map,
    )

    ax.set_xticklabels(
        ax.get_xticklabels(),
        rotation=plot_opts.angle_rotate_xaxis_labels,
        family=plot_opts.plot_font_family,
    )
    ax.set_yticklabels(
        ax.get_yticklabels(),
        rotation=plot_opts.angle_rotate_yaxis_labels,
        family=plot_opts.plot_font_family,
    )
    ax.set_xlabel("Feature Name", family=plot_opts.plot_font_family)
    ax.set_ylabel("Importance", family=plot_opts.plot_font_family)
    ax.set_title(title, family=plot_opts.plot_font_family, wrap=True)

    return fig


def plot_local_shap_importance(
    shap_values: shap.Explainer,
    plot_opts: PlottingOptions,
    num_features_to_plot: int,
    title: str,
) -> Figure:
    """Plot a beeswarm plot of the local SHAP values.

    Args:
        shap_values (shap.Explainer): The SHAP explainer to produce the plot from.
        plot_opts (PlottingOptions): The plotting options.
        num_features_to_plot (int): The number of top features to plot.
        title (str): The plot title.

    Returns:
        Figure: The beeswarm plot of local SHAP values.
    """
    # Plot bee swarm plot
    plt.style.use(plot_opts.plot_colour_scheme)
    fig, ax = plt.subplots(layout="constrained", dpi=plot_opts.dpi)
    ax.set_title(
        title,
        family=plot_opts.plot_font_family,
        wrap=True,
    )
    shap.plots.beeswarm(
        shap_values,
        max_display=num_features_to_plot,
        show=False,
        color=plot_opts.plot_colour_map,
    )
    ax.set_xlabel(ax.get_xlabel(), family=plot_opts.plot_font_family)
    ax.set_ylabel(ax.get_ylabel(), family=plot_opts.plot_font_family)
    ax.set_xticklabels(
        ax.get_xticklabels(),
        family=plot_opts.plot_font_family,
    )
    ax.set_yticklabels(
        ax.get_yticklabels(),
        family=plot_opts.plot_font_family,
    )

    return fig


def plot_global_shap_importance(
    shap_values: pd.DataFrame,
    plot_opts: PlottingOptions,
    num_features_to_plot: int,
    title: str,
) -> Figure:
    """Produce a bar chart of global SHAP values.

    Args:
        shap_values (pd.DataFrame): The `DataFrame` containing the global SHAP values.
        plot_opts (PlottingOptions): The plotting options.
        num_features_to_plot (int): The number of top features to plot.
        title (str): The plot title.

    Returns:
        Figure: The bar chart of global SHAP values.
    """
    # Plot bar chart
    plt.style.use(plot_opts.plot_colour_scheme)
    fig, ax = plt.subplots(layout="constrained", dpi=plot_opts.dpi)
    ax.set_title(
        title,
        family=plot_opts.plot_font_family,
        wrap=True,
    )
    plot_data = (
        shap_values.sort_values(by="SHAP Importance", ascending=False)
        .head(num_features_to_plot)
        .T
    )
    sns.barplot(data=plot_data, fill=True, ax=ax)
    ax.set_xlabel("Feature Name", family=plot_opts.plot_font_family)
    ax.set_ylabel("Abs. SHAP Importance", family=plot_opts.plot_font_family)
    ax.set_xticklabels(
        ax.get_xticklabels(),
        rotation=plot_opts.angle_rotate_xaxis_labels,
        family=plot_opts.plot_font_family,
    )
    ax.set_yticklabels(
        ax.get_yticklabels(),
        rotation=plot_opts.angle_rotate_yaxis_labels,
        family=plot_opts.plot_font_family,
    )

    return fig


def plot_auc_roc(
    y_classes_labels: np.ndarray,
    y_score_probs: np.ndarray,
    model_name: str,
    set_name: str,
    directory: Path,
    plot_opts: PlottingOptions,
) -> None:
    """Plot the ROC curve for a multi-class classification model.

    Args:
        y_classes_labels (numpy.ndarray): The true labels of the classes.
        y_score_probs (numpy.ndarray): The predicted probabilities of the classes.
        model_name (string): The name of the model.
        set_name (string): The name of the set (train or test).
        directory (Path): The directory path to save the plot.
        plot_opts (PlottingOptions): The plotting options.
    """

    num_classes = y_score_probs.shape[1]
    start_index = 1 if num_classes == 2 else 0

    # Set colour scheme
    plt.style.use(plot_opts.plot_colour_scheme)

    for i in range(start_index, num_classes):

        auroc = RocCurveDisplay.from_predictions(
            y_classes_labels[:, i],
            y_score_probs[:, i],
            name=f"Class {i} vs the rest",
            color="darkorange",
            plot_chance_level=True,
        )

        auroc.ax_.set_xlabel(
            "False Positive Rate",
            fontsize=plot_opts.plot_axis_font_size,
            family=plot_opts.plot_font_family,
        )

        auroc.ax_.set_ylabel(
            "True Positive Rate",
            fontsize=plot_opts.plot_axis_font_size,
            family=plot_opts.plot_font_family,
        )

        display_model_name = " ".join(word.capitalize() for word in model_name.split())
        figure_title = f"{display_model_name} {set_name} One-vs-Rest ROC curves:\n {i} Class vs Rest"
        auroc.ax_.set_title(
            figure_title,
            family=plot_opts.plot_font_family,
            fontsize=plot_opts.plot_title_font_size,
            wrap=True,
        )

        auroc.ax_.legend(
            prop={
                "family": plot_opts.plot_font_family,
                "size": plot_opts.plot_axis_tick_size,
            },
            loc="lower right",
        )

        # Save plot with standardized filename
        auroc.figure_.savefig(
            directory / f"{model_name}-{set_name.lower()}-{i}_vs_rest-roc.png",
            dpi=plot_opts.dpi,
        )

        plt.close()


def plot_confusion_matrix(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    model_name: str,
    set_name: str,
    directory: Path,
    plot_opts: PlottingOptions,
) -> None:
    """Plot the confusion matrix for a multi-class or binary classification model.

    Args:
        y_true (np.ndarray): The true labels.
        y_pred (np.ndarray): The predicted labels.
        model_name (str): The name of the model.
        set_name (str): The name of the set (train or test).
        directory (Path): The directory path to save the plot.
        plot_opts (PlottingOptions): The plotting options.
    """
    plt.style.use(plot_opts.plot_colour_scheme)

    # Create title with capitalized model name
    display_model_name = " ".join(word.capitalize() for word in model_name.split())

    # Create confusion matrix display
    disp = ConfusionMatrixDisplay.from_predictions(
        y_true=y_true,
        y_pred=y_pred,
        normalize=None,
        colorbar=False,
        cmap=plot_opts.plot_colour_map,
    )

    title = f"Confusion Matrix - {display_model_name} - {set_name}"
    disp.ax_.set_title(
        title,
        fontsize=plot_opts.plot_title_font_size,
        fontfamily=plot_opts.plot_font_family,
        wrap=True,
    )

    disp.ax_.set_xlabel(
        "Predicted label",
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )
    disp.ax_.set_ylabel(
        "True label",
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
        rotation=plot_opts.angle_rotate_yaxis_labels,
    )

    # Save plot with standardized filename
    disp.figure_.savefig(
        directory / f"{model_name}-{set_name.lower()}-confusion.png", dpi=plot_opts.dpi
    )

    plt.close()
    plt.clf()


def plot_scatter(
    y,
    yp,
    r2: float,
    dependent_variable: str,
    plot_opts: PlottingOptions,
    point_colour: str = "#1f77b4",
    edge_color: str = "#1f77b4",
    point_size: int = 20,
    show_grid: bool = True,
    hue=None,
    style_by: str | None = None,
    palette: str = sns.color_palette("pastel"),
) -> Figure:
    """Create a scatter plot comparing predicted vs actual values.

    Args:
        y (_type_): True y values.
        yp (_type_): Predicted y values.
        r2 (float): R-squared between `y`and `yp`.
        set_name (str): "Train" or "Test".
        dependent_variable (str): The name of the dependent variable.
        model_name (str): Name of the model.
        plot_opts (PlottingOptions): Options for styling the plot.

    Returns:
        Figure: The scatter plot figure
    """

    # Create a scatter plot using Seaborn
    plt.style.use(plot_opts.plot_colour_scheme)
    fig, ax = plt.subplots(
        figsize=(plot_opts.width, plot_opts.height), dpi=plot_opts.dpi
    )
    sns.scatterplot(
        x=y,
        y=yp,
        ax=ax,
        color=point_colour,
        edgecolor=edge_color,
        s=point_size,
        hue=hue,
        style=style_by,
        palette=palette,
    )

    # Add the best fit line
    ax.plot([y.min(), y.max()], [y.min(), y.max()], "k--", lw=2, label="Best fit")

    # Set labels and title
    x_label = (
        plot_opts.xaxis_label
        if plot_opts.xaxis_label
        else "Measured " + dependent_variable
    )
    ax.set_xlabel(
        x_label,
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )

    y_label = (
        plot_opts.yaxis_label
        if plot_opts.yaxis_label
        else "Predicted " + dependent_variable
    )

    ax.set_ylabel(
        y_label,
        fontsize=plot_opts.plot_axis_font_size,
        family=plot_opts.plot_font_family,
    )

    figure_title = plot_opts.plot_title if plot_opts.plot_title else "Parity Plot"

    ax.set_title(
        figure_title,
        fontsize=plot_opts.plot_title_font_size,
        family=plot_opts.plot_font_family,
        wrap=True,
    )

    if r2 is not None:
        # Add legend with R2 value
        if isinstance(r2, dict) and "value" in r2:
            r2_value = r2["value"]
        else:
            r2_value = r2
        legend = f"R²: {r2_value:.3f}"
        ax.plot([], [], " ", label=legend)  # Add empty line for R2 in legend
        ax.legend(
            prop={
                "family": plot_opts.plot_font_family,
                "size": plot_opts.plot_axis_tick_size,
            },
            loc="upper left",
        )

    # Add grid
    ax.grid(visible=show_grid, axis="both", linestyle="--", alpha=0.6)

    # Adjust layout
    plt.tight_layout()

    return fig


def plot_beta_coefficients(
    coefficients: np.ndarray,
    feature_names: list,
    plot_opts: PlottingOptions,
    model_name: str,
    dependent_variable: str | None = None,
    standard_errors: np.ndarray | None = None,
    is_classification: bool = False,
) -> Figure:
    """Create a bar plot of model coefficients with different colors for positive/negative values.

    Args:
        coefficients (np.ndarray): The model coefficients. For logistic regression, uses first class coefficients
        feature_names (list): Names of the features
        plot_opts (PlottingOptions): Plot styling options
        model_name (str): Name of the model for the plot title
        dependent_variable (str | None, optional): Name of the dependent variable. Defaults to None.
        standard_errors (np.ndarray | None, optional): Standard errors of coefficients. Defaults to None.
        is_classification (bool, optional): Whether this is a classification model. Defaults to False.

    Returns:
        Figure: The coefficient plot
    """
    plt.style.use(plot_opts.plot_colour_scheme)

    # Handle different coefficient shapes
    if len(coefficients.shape) > 1:
        coefficients = coefficients[0]

    # Ensure coefficients and feature names match in length
    if len(coefficients) != len(feature_names):
        raise ValueError(
            f"Number of coefficients ({len(coefficients)}) does not match "
            f"number of feature names ({len(feature_names)})"
        )

    # Sort coefficients by their values
    sorted_indices = np.argsort(coefficients)
    coefficients = coefficients[sorted_indices]
    feature_names = [feature_names[i] for i in sorted_indices]
    if standard_errors is not None:
        standard_errors = standard_errors[sorted_indices]

    # Calculate figure dimensions based on number of coefficients
    n_coef = len(coefficients)
    base_height = 4  # minimum height in inches
    height_per_coef = 0.25  # additional height per coefficient
    fig_height = min(base_height + height_per_coef * n_coef, 20)  # cap at 20 inches

    # Only scale feature names font size based on number of coefficients
    font_scale = max(0.5, min(1.0, 25 / n_coef))  # scale between 0.5 and 1.0
    axis_font_size = int(plot_opts.plot_axis_font_size * font_scale)

    # Create figure with calculated dimensions
    fig, ax = plt.subplots(figsize=(10, fig_height), dpi=plot_opts.dpi)

    # Get colors from the palette
    palette = sns.color_palette(plot_opts.plot_colour_map)
    negative_color = palette[0]  # First color for negative values
    positive_color = palette[-1]  # Last color for positive values

    # Create color list based on coefficient signs
    colors = [negative_color if c < 0 else positive_color for c in coefficients]

    # Plot horizontal bars
    y_pos = np.arange(len(coefficients))
    bars = ax.barh(y_pos, coefficients, color=colors)

    # Add error bars if provided
    if standard_errors is not None:
        error_positions = np.array(
            [
                (
                    bar.get_width()
                    if bar.get_width() >= 0
                    else bar.get_width() + standard_errors[i]
                )
                for i, bar in enumerate(bars)
            ]
        )
        ax.errorbar(
            error_positions,
            y_pos,
            xerr=standard_errors,
            fmt="none",
            color="black",
            capsize=5,
            zorder=2,
        )

    # Customize axes with scaled font size only for feature names
    ax.set_yticks(y_pos)
    ax.set_yticklabels(
        feature_names,
        family=plot_opts.plot_font_family,
        fontsize=axis_font_size,
    )

    # Create title
    if model_name.lower() == "linear model":
        display_model_name = (
            "Logistic Regression" if is_classification else "Linear Regression"
        )
    else:
        display_model_name = " ".join(word.capitalize() for word in model_name.split())

    title = f"Beta Coefficients - {display_model_name}"
    if dependent_variable:
        title += f"\nDependent Variable: {dependent_variable}"

    # Use original title font size
    ax.set_title(
        title,
        family=plot_opts.plot_font_family,
        fontsize=plot_opts.plot_title_font_size,
        wrap=True,
    )

    # Add gridlines
    ax.grid(True, axis="x", linestyle="--", alpha=0.3, zorder=0)

    # Add zero line
    ax.axvline(x=0, color="black", linestyle="-", linewidth=0.5, zorder=1)

    # Adjust layout
    plt.tight_layout()

    return fig


def plot_permutation_importance(
    df: pd.DataFrame,
    plot_opts: PlottingOptions,
    n_features: int,
    title: str,
) -> Figure:
    """Plot a bar chart of the top n features in the feature importance dataframe,
    with the given title and styled with the given options.

    Args:
        df (pd.DataFrame): The dataframe containing the permutation importance.
        plot_opts (PlottingOptions): The options for how to configure the plot.
        n_features (int): The top number of features to plot.
        title (str): The title of the plot.

    Returns:
        Figure: The bar chart of the top n features.
    """

    plt.style.use(plot_opts.plot_colour_scheme)
    fig, ax = plt.subplots(layout="constrained", dpi=plot_opts.dpi)

    top_features = (
        df.sort_values(by="Permutation Importance", ascending=False).head(n_features).T
    )
    sns.barplot(top_features, ax=ax)

    ax.set_xticklabels(
        ax.get_xticklabels(),
        rotation=plot_opts.angle_rotate_xaxis_labels,
        family=plot_opts.plot_font_family,
    )
    ax.set_yticklabels(
        ax.get_yticklabels(),
        rotation=plot_opts.angle_rotate_yaxis_labels,
        family=plot_opts.plot_font_family,
    )
    ax.set_xlabel("Feature", family=plot_opts.plot_font_family)
    ax.set_ylabel("Importance", family=plot_opts.plot_font_family)
    ax.set_title(
        title,
        family=plot_opts.plot_font_family,
        wrap=True,
    )

    return fig


def plot_bar_chart(
    df: pd.DataFrame,
    sort_key: Any,
    plot_opts: PlottingOptions,
    title: str,
    x_label: str,
    y_label: str,
    n_features: int = 10,
    error_bars: pd.DataFrame | None = None,
) -> Figure:
    """Plot a bar chart of the top n features from the given dataframe.

    Args:
        df (pd.DataFrame): The data to be plotted.
        plot_opts (PlottingOptions): The options for styling the plot.
        sort_key (str): The key by which to sort the data. This can be the name of a column.
        title (str): The title of the plot.
        x_label (str): The label for the X axis.
        y_label (str): The label for the Y axis.
        n_features (int, optional): The top number of featurs to plot. Defaults to 10.
        error_bars (pd.DataFrame | None, optional): Error bars for the plot. Defaults to None.
        directory (Path | None, optional): The directory to save the plot. Defaults to None.
        model_name (str | None, optional): The name of the model. Defaults to None.
        set_name (str | None, optional): The name of the set (train or test). Defaults to None.

    Returns:
        Figure: The bar chart of the top n features.
    """

    plt.style.use(plot_opts.plot_colour_scheme)
    fig, ax = plt.subplots(layout="constrained", dpi=plot_opts.dpi)

    # Get top features
    top_df = df.sort_values(by=sort_key, ascending=False).head(n_features)
    x = top_df.index.tolist()
    y = top_df[sort_key].values

    # If error bars provided, align them
    yerr = None
    if error_bars is not None and sort_key in error_bars.columns:
        yerr = error_bars.loc[x, sort_key].values

    # Plot with error bars
    ax.bar(x, y, yerr=yerr, capsize=5)

    # Label formatting
    ax.set_xticklabels(
        x,
        rotation=plot_opts.angle_rotate_xaxis_labels,
        family=plot_opts.plot_font_family,
    )
    ax.set_yticklabels(
        ax.get_yticks(),
        rotation=plot_opts.angle_rotate_yaxis_labels,
        family=plot_opts.plot_font_family,
    )
    ax.set_xlabel(x_label, family=plot_opts.plot_font_family)
    ax.set_ylabel(y_label, family=plot_opts.plot_font_family)
    ax.yaxis.set_major_formatter(FormatStrFormatter("%.1f"))
    ax.set_title(title, family=plot_opts.plot_font_family, wrap=True)

    return fig



from pathlib import Path

import streamlit as st

from helix.components.configuration import display_options
from helix.components.experiments import experiment_selector
from helix.components.forms.forms_preprocessing import preprocessing_opts_form
from helix.components.images.logos import sidebar_logo
from helix.components.preprocessing import original_view, preprocessed_view
from helix.options.data import DataOptions
from helix.options.enums import DataPreprocessingStateKeys, ExecutionStateKeys
from helix.options.file_paths import (
    data_options_path,
    data_preprocessing_options_path,
    execution_options_path,
    helix_experiments_base_dir,
    plot_options_path,
    preprocessed_data_path,
)
from helix.options.preprocessing import PreprocessingOptions
from helix.services.configuration import (
    load_data_options,
    load_data_preprocessing_options,
    load_execution_options,
    load_plot_options,
    save_options,
)
from helix.services.data import read_data, save_data
from helix.services.experiments import get_experiments
from helix.services.preprocessing import find_non_numeric_columns, run_preprocessing
from helix.utils.logging_utils import Logger, close_logger


def build_config() -> PreprocessingOptions:
    """
    Build the configuration object for preprocessing.
    """

    preprocessing_options = PreprocessingOptions(
        feature_selection_methods={
            DataPreprocessingStateKeys.VarianceThreshold: st.session_state[
                DataPreprocessingStateKeys.VarianceThreshold
            ],
            DataPreprocessingStateKeys.CorrelationThreshold: st.session_state[
                DataPreprocessingStateKeys.CorrelationThreshold
            ],
            DataPreprocessingStateKeys.LassoFeatureSelection: st.session_state[
                DataPreprocessingStateKeys.LassoFeatureSelection
            ],
        },
        variance_threshold=st.session_state[
            DataPreprocessingStateKeys.ThresholdVariance
        ],
        correlation_threshold=st.session_state[
            DataPreprocessingStateKeys.ThresholdCorrelation
        ],
        lasso_regularisation_term=st.session_state[
            DataPreprocessingStateKeys.RegularisationTerm
        ],
        independent_variable_normalisation=st.session_state[
            DataPreprocessingStateKeys.IndependentNormalisation
        ].lower(),
        dependent_variable_transformation=st.session_state[
            DataPreprocessingStateKeys.DependentNormalisation
        ].lower(),
    )
    return preprocessing_options


st.set_page_config(
    page_title="Data Preprocessing",
    page_icon=sidebar_logo(),
)

sidebar_logo()

st.header("Data Preprocessing")
st.write(
    """
    Here you can preprocess your data before running machine learning models. This includes feature selection and scalling of variables.
    """
)

choices = get_experiments()
experiment_name = experiment_selector(choices)
helix_base_dir = helix_experiments_base_dir()


def validate_data(data, id_col: str | None) -> tuple[list, bool]:
    """Validate data for preprocessing.

    Args:
        data: The input data to validate
        id_col: The name of the ID column

    Returns:
        tuple containing list of non-numeric columns and whether y has non-numeric values
    """
    non_numeric = find_non_numeric_columns(data.iloc[:, :-1])
    if id_col in non_numeric:
        non_numeric.remove(id_col)

    if non_numeric:
        st.warning(
            f"The following columns contain non-numeric values: {', '.join(non_numeric)}. These will be eliminated."
        )
    else:
        st.success("All the independent variable columns are numeric.")

    non_numeric_y = find_non_numeric_columns(data.iloc[:, -1])
    if non_numeric_y:
        st.warning(
            "The dependent variable contains non-numeric values. This will be transformed to allow for training."
        )

    return non_numeric, bool(non_numeric_y)


def run_preprocessing_pipeline(
    data,
    config,
    experiment_dir: Path,
    data_opts: DataOptions,
    path_to_data_opts,
    path_to_preproc_opts,
    logger,
) -> None:
    """Run the preprocessing pipeline and save results.

    Args:
        data: Input data to preprocess
        config: Preprocessing configuration
        experiment_dir: Path to experiment directory
        data_opts: Data options
        path_to_data_opts: Path to data options file
        path_to_preproc_opts: Path to preprocessing options file
        logger: Logger instance
    """
    processed_data = run_preprocessing(
        data, experiment_dir, config, id_col=data_opts.id_column
    )

    path_to_preprocessed_data = preprocessed_data_path(
        Path(data_opts.data_path).name,
        experiment_dir,
    )

    save_data(path_to_preprocessed_data, processed_data, logger)

    # Update data opts to point to the pre-processed data
    data_opts.data_path = str(path_to_preprocessed_data)
    save_options(path_to_data_opts, data_opts)

    # Update config to show preprocessing is complete
    config.data_is_preprocessed = True
    save_options(path_to_preproc_opts, config)

    st.success("Data Preprocessing Complete")
    st.header(f"Preprocessed Data ({processed_data.shape[1]} columns)")
    preprocessed_view(processed_data)


if experiment_name:
    logger_instance = Logger()
    logger = logger_instance.make_logger()

    try:
        st.session_state[ExecutionStateKeys.ExperimentName] = experiment_name
        display_options(helix_base_dir / experiment_name)

        path_to_plot_opts = plot_options_path(helix_base_dir / experiment_name)
        path_to_data_opts = data_options_path(helix_base_dir / experiment_name)
        data_opts = load_data_options(path_to_data_opts)

        path_to_preproc_opts = data_preprocessing_options_path(
            helix_base_dir / experiment_name
        )

        exec_opts_path = execution_options_path(helix_base_dir / experiment_name)
        exec_opts = load_execution_options(exec_opts_path)

        # Check preprocessing status
        data_is_preprocessed = False
        if path_to_preproc_opts.exists():
            preproc_opts = load_data_preprocessing_options(path_to_preproc_opts)
            data_is_preprocessed = preproc_opts.data_is_preprocessed

        if data_is_preprocessed:
            st.warning(
                "Your data are already preprocessed. Would you like to start again?"
            )
            preproc_again = st.checkbox("Redo preprocessing", value=False)
            if preproc_again:
                st.cache_data.clear()
        else:
            preproc_again = True

        if not preproc_again:
            data = read_data(Path(data_opts.data_path), logger)
            preprocessed_view(data)
        else:
            # remove preprocessed suffix to point to original data file
            data_opts.data_path = data_opts.data_path.replace("_preprocessed", "")
            data = read_data(Path(data_opts.data_path), logger)

        # Validate data
        validate_data(data, id_col=data_opts.id_column)
        plot_opt = load_plot_options(path_to_plot_opts)
        original_view(data)
        preprocessing_opts_form(data, exec_opts.problem_type)

        if st.button("Run Data Preprocessing", type="primary"):
            config = build_config()
            run_preprocessing_pipeline(
                data,
                config,
                helix_base_dir / experiment_name,
                data_opts,
                path_to_data_opts,
                path_to_preproc_opts,
                logger,
            )

    except ValueError as e:
        st.error(str(e), icon="🔥")
    except Exception:
        st.error("Something went wrong.", icon="🔥")
    finally:
        close_logger(logger_instance, logger)



from helix.options.choices.metrics import CLASSIFICATION_METRICS, REGRESSION_METRICS
from helix.options.enums import ProblemTypes


def get_metrics(problem_type: ProblemTypes, logger: object = None) -> dict:
    """Get the metrics functions for a given problem type.

    For classification:
    - Accuracy
    - F1
    - Precision
    - Recall
    - ROC AUC

    For Regression
    - R2
    - MAE
    - RMSE

    Args:
        problem_type (ProblemTypes): Where the problem is classification or regression.
        logger (object, optional): The logger. Defaults to None.

    Raises:
        ValueError: When you give an incorrect problem type.

    Returns:
        dict: A `dict` of score names and functions.
    """
    if problem_type.lower() == ProblemTypes.Classification:
        metrics = CLASSIFICATION_METRICS
    elif problem_type.lower() == ProblemTypes.Regression:
        metrics = REGRESSION_METRICS
    else:
        raise ValueError(f"Problem type {problem_type} not recognized")

    logger.info(f"Using metrics: {list(metrics.keys())}")
    return metrics


def find_mean_model_index(
    full_metrics: dict, aggregated_metrics: dict, metric_name: str
) -> int:
    """
    Find the index of the model with the mean of the metric.
    """

    for model_name, stats in aggregated_metrics.items():
        # Extract the mean metric for the test set
        mean_metric_test = stats["test"][metric_name]["mean"]

        # Find the bootstrap index closest to the mean metric
        dif = float("inf")
        closest_index = -1
        for i, bootstrap in enumerate(full_metrics[model_name]):
            metric_value = bootstrap[metric_name]["test"]["value"]
            current_dif = abs(metric_value - mean_metric_test)
            if current_dif < dif:
                dif = current_dif
                closest_index = i

    return closest_index



from sklearn.metrics import (
    accuracy_score,
    f1_score,
    mean_absolute_error,
    precision_score,
    r2_score,
    recall_score,
    roc_auc_score,
    root_mean_squared_error,
)

CLASSIFICATION_METRICS = {
    "accuracy": accuracy_score,
    "f1_score": f1_score,
    "precision_score": precision_score,
    "recall_score": recall_score,
    "roc_auc_score": roc_auc_score,
}
REGRESSION_METRICS = {
    "MAE": mean_absolute_error,
    "RMSE": root_mean_squared_error,
    "R2": r2_score,
}



from pathlib import Path

import streamlit as st

from helix.options.enums import ExecutionStateKeys, ViewExperimentKeys


def experiment_selector(options: list) -> str:
    """Select

    Args:
        options (list): The list of experiment names to choose from.

    Returns:
        str: The name of the experiment on disk.
    """

    return st.selectbox(
        "Select an experiment",
        options=options,
        index=None,
        placeholder="Experiment name",
        key=ViewExperimentKeys.ExperimentName,
    )


def model_selector(options: list, gui_text: str, placeholder: str, key: str) -> Path:
    """Select a model or models for their intended use. This function creates a multiselect widget
    to allow the user to select multiple models to use for FI pipeline or to Predict on new datasets.

    Args:
        options (list): The list of model names to choose from.
        gui_text (str): The text to display above the widget indicating what the models are meant to be used for.
        placeholder (str): The placeholder to show on the widget before the user does the selection.
        key (str): the key to use to save the information of the widget in the state session.

    Returns:
        Path: The path to the model on disk.
    """

    return st.multiselect(
        gui_text,
        options=options,
        default=None,
        placeholder=placeholder,
        key=key,
    )


def data_selector(options: list) -> Path:
    """Select a model or models to explain. This function creates a multiselect widget
    to allow the user to select multiple models to explain using the FI pipeline.

    Args:
        options (list): The list of model names to choose from.

    Returns:
        Path: The path to the model on disk.
    """

    return st.selectbox(
        "Select a dataset to explain",
        options=options,
        index=None,
        placeholder="Dataset to explain",
        key=ExecutionStateKeys.UploadedFileName,
    )



import warnings
from typing import Any

import pandas as pd
import shap
from lime.lime_tabular import LimeTabularExplainer
from sklearn.base import is_classifier

from helix.options.enums import ProblemTypes
from helix.utils.logging_utils import Logger


def calculate_lime_values(
    model, X: pd.DataFrame, problem_type: ProblemTypes, logger: Logger
) -> pd.DataFrame:
    """Calculate LIME values for a given model and dataset.

    Args:
        model: The model.
        X (pd.DataFrame): The dataset.
        problem_type (ProblemTypes): The problem type.
        logger (Logger): The logger.

    Returns:
        pd.DataFrame: The LIME values.
    """
    logger.info(f"Calculating LIME Importance for {model.__class__.__name__} model..")

    # Suppress all warnings
    warnings.filterwarnings("ignore")
    explainer = LimeTabularExplainer(X.to_numpy(), mode=problem_type)

    coefficients = []

    for i in range(X.shape[0]):
        if is_classifier(model):
            explanation = explainer.explain_instance(
                X.iloc[i, :], model.predict_proba, num_features=X.shape[1]
            )
        else:
            explanation = explainer.explain_instance(
                X.iloc[i, :], model.predict, num_features=X.shape[1]
            )

        coefficients.append([item[-1] for item in explanation.local_exp[1]])

    lr_lime_values = pd.DataFrame(coefficients, columns=X.columns, index=X.index)

    # TODO: scale coefficients between 0 and +1 (low to high impact)

    logger.info("LIME Importance Analysis Completed..")

    return lr_lime_values


def calculate_local_shap_values(
    model,
    X: pd.DataFrame,
    logger: Logger,
) -> tuple[pd.DataFrame, Any]:
    """Calculate local SHAP values for a given model and dataset.

    Args:
        model: Model object.
        X (pd.DataFrame): The dataset.
        logger (Logger): The logger.

    Returns:
        tuple[pd.DataFrame, Any]: SHAP dataframe and SHAP values.
    """
    logger.info(f"Calculating SHAP Importance for {model.__class__.__name__} model..")

    explainer = shap.Explainer(model.predict, X)

    shap_values = explainer(X)

    shap_df = pd.DataFrame(shap_values.values, columns=X.columns, index=X.index)
    # TODO: scale coefficients between 0 and +1 (low to high impact)

    logger.info("SHAP Importance Analysis Completed..")

    # Return the DataFrame
    return shap_df, shap_values



import os

from helix.machine_learning.learner import GridSearchLearner, Learner
from helix.options.data import DataOptions
from helix.options.execution import ExecutionOptions
from helix.options.ml import MachineLearningOptions
from helix.options.plotting import PlottingOptions
from helix.services.data import DataBuilder
from helix.services.machine_learning.results import save_actual_pred_plots
from helix.utils.logging_utils import Logger


def run(
    ml_opts: MachineLearningOptions,
    data_opts: DataOptions,
    plot_opts: PlottingOptions,
    data: DataBuilder,
    exec_opts: ExecutionOptions,
    logger: Logger,
    n_cpus: int,
) -> None:
    """
    Run the ML training pipeline
    """

    if ml_opts.use_hyperparam_search:
        learner = GridSearchLearner(
            model_types=ml_opts.model_types,
            problem_type=exec_opts.problem_type,
            data_split=data_opts.data_split,
            logger=logger,
            n_cpus=n_cpus,
        )
    else:
        learner = Learner(
            model_types=ml_opts.model_types,
            problem_type=exec_opts.problem_type,
            data_split=data_opts.data_split,
            logger=logger,
            n_cpus=n_cpus,
        )
    res, metrics_full, metrics_mean_std, trained_models = learner.fit(data)
    logger.info(f"Performance Metric Statistics: {os.linesep}{metrics_mean_std}")
    if ml_opts.save_actual_pred_plots:
        save_actual_pred_plots(
            data=data,
            ml_results=res,
            logger=logger,
            ml_metric_results=metrics_full,
            ml_metric_results_stats=metrics_mean_std,
            exec_opts=exec_opts,
            plot_opts=plot_opts,
            ml_opts=ml_opts,
            trained_models=trained_models,
        )

    return trained_models, metrics_full, metrics_mean_std



from dataclasses import dataclass


@dataclass
class PreprocessingOptions:
    feature_selection_methods: dict
    variance_threshold: float
    correlation_threshold: float
    lasso_regularisation_term: float
    independent_variable_normalisation: str = "none"
    dependent_variable_transformation: str = "none"
    data_is_preprocessed: bool = False



from copy import deepcopy

import pandas as pd

from helix.options.enums import ProblemTypes
from helix.options.execution import ExecutionOptions
from helix.options.fi import FeatureImportanceOptions
from helix.options.fuzzy import FuzzyOptions
from helix.options.plotting import PlottingOptions
from helix.services.data import TabularData
from helix.services.feature_importance.results import (
    save_fuzzy_sets_plots,
    save_importance_results,
    save_target_clusters_plots,
)
from helix.utils.logging_utils import Logger


class Fuzzy:
    """
    Fuzzy class to interpret synergy of importance between features within context.

    """

    def __init__(
        self,
        fuzzy_opt: FuzzyOptions,
        fi_opt: FeatureImportanceOptions,
        exec_opt: ExecutionOptions,
        plot_opt: PlottingOptions,
        # ml_opt: argparse.Namespace,
        logger: Logger | None = None,
    ) -> None:
        self._fuzzy_opt = fuzzy_opt
        self._fi_opt = fi_opt
        self._exec_opt = exec_opt
        self._plot_opt = plot_opt
        self._logger = logger
        self._local_importance_methods = self._fi_opt.local_importance_methods
        self.importance_type = "local"  # local feature importance

    def interpret(self, ensemble_results: dict, local_results: dict, data: TabularData):
        """Interpret the model results using the selected feature importance methods and ensemble methods.

        Args:
            ensemble_results (dict): Ensemble FI results used to select top features.
            local_results (dict): Local FI results.
            data (TabularData): The data used in training.

        Returns:
            dict: Dictionary of feature importance results.
        """

        # create a copy of the data - select first fold of the data
        X_train, X_test = deepcopy(data.X_train[0]), deepcopy(data.X_test[0])
        self._logger.info("-------- Start of fuzzy interpretation logging--------")
        # Step 1: fuzzy feature selection to select top features for fuzzy interpretation
        if self._fuzzy_opt.fuzzy_feature_selection:
            # Select top features for fuzzy interpretation
            try:
                topfeatures = self._select_features(ensemble_results["Majority Vote"])
                X_train = X_train[topfeatures]
                X_test = X_test[topfeatures]
            except Exception as e:
                self._logger.error(f"Error in fuzzy feature selection: {e}")
                try:
                    # Use Mean ensemble results
                    topfeatures = self._select_features(ensemble_results["Mean"])
                    X_train = X_train[topfeatures]
                    X_test = X_test[topfeatures]
                except Exception as e:
                    self._logger.error(f"Error in fuzzy feature selection: {e}")

            # if error occurs, use Mean ensemble results

        # Step 2: Assign granularity to features e.g. low, medium, high categories
        if self._fuzzy_opt.granular_features:
            X_train = self._fuzzy_granularity(X_train)
            X_test = self._fuzzy_granularity(X_test)

        # Step 3.1: Convert local FI results to a dataframe
        master_importance_df = pd.DataFrame()
        for _, feature_importance in local_results.items():
            for _, result in feature_importance.items():
                master_importance_df = pd.concat([master_importance_df, result], axis=0)
        master_importance_df.reset_index(drop=True, inplace=True)

        # Step 3.2: Extract fuzzy rules from master dataframe
        fuzzy_rules_df = self._fuzzy_rule_extraction(master_importance_df)
        save_importance_results(
            feature_importance_df=fuzzy_rules_df,
            model_type=None,
            importance_type="fuzzy",
            feature_importance_type="top contextual rules",
            experiment_name=self._exec_opt.experiment_name,
            fi_opt=self._fi_opt,
            plot_opt=self._plot_opt,
            logger=self._logger,
        )

        # Step 4: Identify the synergy of important features by context (e.g. target category:low, medium, high)
        df_contextual_rules = self._contextual_synergy_analysis(fuzzy_rules_df)
        save_importance_results(
            feature_importance_df=df_contextual_rules,
            model_type=None,
            importance_type="fuzzy",
            feature_importance_type="top contextual rules",
            experiment_name=self._exec_opt.experiment_name,
            fi_opt=self._fi_opt,
            plot_opt=self._plot_opt,
            logger=self._logger,
        )

        self._logger.info("-------- End of fuzzy interpretation logging--------")

        return df_contextual_rules

    def _select_features(self, ensemble_results):
        """
        Select top features from ensemble feature importance. This can be either
        majority vote-based importance or mean-based importance
        Parameters:
            ensemble_results: Dictionary of feature importance results.
        Returns:
            list: List of top features.
        """
        self._logger.info(
            f"Selecting top {self._fuzzy_opt.number_fuzzy_features} features..."
        )
        fi = ensemble_results.sort_values(by=0, ascending=False)
        # Select top n features for fuzzy interpretation
        topfeatures = fi.index[: self._fuzzy_opt.number_fuzzy_features].tolist()
        return topfeatures

    def _fuzzy_granularity(self, X):
        """
        Generate membership functions for each feature to split them into 3 granularities:
        "small", "moderate", "large". Then use these membership functions to determine
        degree membership for each granularity.

        The Z function is used to determine degree membership of the "small" granularity.

        The triangular function is used to determine degree membership of the "moderate"
        granularity.

        The S function is used to determine degree membership of the "large" granularity.

        Parameters:
            X (pd.DataFrame): Features.
        Returns:
            pd.DataFrame: Features with granularity.
        """
        import warnings

        import numpy as np
        import skfuzzy as fuzz

        # Suppress all warnings
        warnings.filterwarnings("ignore")
        self._logger.info("Assigning granularity to features...")
        # find interquartile values for each feature
        df_top_qtl = X.quantile([0, 0.25, 0.5, 0.75, 1])
        # Create membership functions based on interquartile values for each feature
        membership_functions = {}
        universe = {}
        for feature in X.columns:

            # Define the universe for each feature
            universe[feature] = np.linspace(X[feature].min(), X[feature].max(), 100)

            # Define membership functions
            # features with less than 3 unique values
            if len(X[feature].unique()) < 3:
                # print('Feature with only 2 values')
                # print(feature)
                # print(df_top_qtl[feature])
                low_mf = fuzz.trimf(
                    universe[feature],
                    [
                        df_top_qtl[feature][0.00],
                        df_top_qtl[feature][0.00],
                        df_top_qtl[feature][0.00],
                    ],
                )
                medium_mf = fuzz.trimf(
                    universe[feature],
                    [
                        (df_top_qtl[feature][0.00] + df_top_qtl[feature][1.00]) / 2,
                        (df_top_qtl[feature][0.00] + df_top_qtl[feature][1.00]) / 2,
                        (df_top_qtl[feature][0.00] + df_top_qtl[feature][1.00]) / 2,
                    ],
                )
                high_mf = fuzz.trimf(
                    universe[feature],
                    [
                        df_top_qtl[feature][1.00],
                        df_top_qtl[feature][1.00],
                        df_top_qtl[feature][1.00],
                    ],
                )

            # Highly skewed features
            elif df_top_qtl[feature][0.00] == df_top_qtl[feature][0.50]:
                low_mf = fuzz.trimf(
                    universe[feature],
                    [
                        df_top_qtl[feature][0.00],
                        df_top_qtl[feature][0.50],
                        df_top_qtl[feature][0.75],
                    ],
                )
                medium_mf = fuzz.trimf(
                    universe[feature],
                    [
                        df_top_qtl[feature][0.50],
                        df_top_qtl[feature][0.75],
                        df_top_qtl[feature][1.00],
                    ],
                )
                high_mf = fuzz.smf(
                    universe[feature],
                    df_top_qtl[feature][0.75],
                    df_top_qtl[feature][1.00],
                )

            else:
                low_mf = fuzz.zmf(
                    universe[feature],
                    df_top_qtl[feature][0.00],
                    df_top_qtl[feature][0.50],
                )
                medium_mf = fuzz.trimf(
                    universe[feature],
                    [
                        df_top_qtl[feature][0.25],
                        df_top_qtl[feature][0.50],
                        df_top_qtl[feature][0.75],
                    ],
                )
                high_mf = fuzz.smf(
                    universe[feature],
                    df_top_qtl[feature][0.50],
                    df_top_qtl[feature][1.00],
                )

            membership_functions[feature] = {
                "small": low_mf,
                "moderate": medium_mf,
                "large": high_mf,
            }
        if self._fuzzy_opt.save_fuzzy_set_plots:
            save_fuzzy_sets_plots(
                universe=universe,
                membership_functions=membership_functions,
                x_cols=X.columns,
                exec_opt=self._exec_opt,
                plot_opt=self._plot_opt,
                logger=self._logger,
            )

        # Create granular features using membership values
        new_df_features = []
        for feature in X.columns:
            X.loc[:, f"{feature}_small"] = fuzz.interp_membership(
                universe[feature], membership_functions[feature]["small"], X[feature]
            )
            new_df_features.append(f"{feature}_small")
            X.loc[:, f"{feature}_mod"] = fuzz.interp_membership(
                universe[feature], membership_functions[feature]["moderate"], X[feature]
            )
            new_df_features.append(f"{feature}_mod")
            X.loc[:, f"{feature}_large"] = fuzz.interp_membership(
                universe[feature], membership_functions[feature]["large"], X[feature]
            )
            new_df_features.append(f"{feature}_large")
        X = X[new_df_features]

        return X

    def _fuzzyset_selection(self, uni, mf1, mf2, mf3, val):
        """
        Select fuzzy set with highest membership value.
        Parameters:
            uni (np.array): Universe.
            mf1 (np.array): Low membership function
            mf2 (np.array): Moderate membership function
            mf3 (np.array): High membership function
            val (float): Value.
        Returns:
            str: Fuzzy set with highest membership value
        """
        import skfuzzy as fuzz

        mf_values = []
        # Calculate membership values for each fuzzy set
        mf_values.append(fuzz.interp_membership(uni, mf1, val))

        mf_values.append(fuzz.interp_membership(uni, mf2, val))

        mf_values.append(fuzz.interp_membership(uni, mf3, val))

        # Select fuzzy set with highest membership value
        # if multiple fuzzy sets have the same membership value, the first one is selected

        index_of_max = mf_values.index(max(mf_values))

        # Return fuzzy set
        match index_of_max:
            case 0:
                return "small"
            case 1:
                return "moderate"
            case 2:
                return "large"

    def _fuzzy_rule_extraction(self, df):
        """
        Cluster the local feature importance data into the user-defined clusters and then
        assign the "small", "moderate" and "large" granularities to features within those
        clusters.

        Parameters:
            df (Dataframe): master dataframe of feature importances from local feature importance methods and ML models.
        Returns:
            pd.DataFrame: Features with fuzzy rules.
        """
        import numpy as np
        import skfuzzy as fuzz

        self._logger.info("Extracting fuzzy rules...")
        if self._exec_opt.problem_type == ProblemTypes.Regression:
            target = np.array(df[df.columns[-1]])
            centers, membership_matrix, _, _, _, _, _ = fuzz.cluster.cmeans(
                data=target.reshape(1, -1),
                c=self._fuzzy_opt.number_clusters,
                m=2,  # Fuzziness parameter
                error=0.005,
                maxiter=1000,
            )
            # Determine the primary cluster assignment for each data point
            primary_cluster_assignment = np.argmax(membership_matrix, axis=0)

            # Calculate the average value for each cluster
            cluster_numbers = np.unique(primary_cluster_assignment).tolist()

            cluster_averages = [
                np.mean(target[primary_cluster_assignment == i])
                for i in range(self._fuzzy_opt.number_clusters)
            ]

            # Replace cluser numbers by cluster names based on their average values
            cluster_names = self._fuzzy_opt.cluster_names

            # Create a list of tuples where each tuple contains a cluster number and its corresponding average
            clusters = list(zip(cluster_numbers, cluster_averages))

            # Sort this list by the averages
            clusters.sort(key=lambda x: x[1])

            # Create a dictionary where the keys are the cluster numbers and the values are the cluster names
            cluster_mapping = {
                cluster_num[0]: cluster_name
                for cluster_num, cluster_name in zip(clusters, cluster_names)
            }

            # replace the cluster numbers in the list with the corresponding cluster names
            primary_cluster_assignment = [
                cluster_mapping[cluster_num]
                for cluster_num in primary_cluster_assignment
            ]

            # TODO: save the plot of the range of cluster assignments
            # create new dataframe with target values and cluster assignments
            df_cluster = pd.DataFrame(
                {
                    "target": df.loc[:, df.columns[-1]].to_list(),
                    "cluster": primary_cluster_assignment,
                }
            )

            if self._fuzzy_opt.save_fuzzy_set_plots:
                save_target_clusters_plots(
                    df_cluster=df_cluster,
                    exec_opt=self._exec_opt,
                    plot_opt=self._plot_opt,
                    logger=self._logger,
                )
            # Assign labels to target
            df.loc[:, df.columns[-1]] = primary_cluster_assignment

        # Create membership functions based on interquartile values for each feature
        membership_functions = {}
        universe = {}
        for feature in df.columns[:-1]:
            # Define the universe for each feature
            universe[feature] = np.linspace(df[feature].min(), df[feature].max(), 100)

            # Define membership functions
            low_mf = fuzz.zmf(universe[feature], 0.00, 0.5)
            medium_mf = fuzz.trimf(universe[feature], [0.25, 0.5, 0.75])
            high_mf = fuzz.smf(universe[feature], 0.5, 1.00)

            membership_functions[feature] = {
                "small": low_mf,
                "moderate": medium_mf,
                "large": high_mf,
            }

        # Create fuzzy rules
        fuzzy_rules = []

        # Loop through each row in the dataframe and extract fuzzy rules
        for i, _ in df.iterrows():
            df_instance = {}  # Dictionary to store observation values
            fuzzy_sets = {}  # Dictionary to store fuzzy sets
            for feature in df.columns[:-1]:
                df_instance[feature] = df.loc[i, feature]
            # Extract fuzzy set for each feature
            for feature in df.columns[:-1]:
                fuzzy_sets[feature] = self._fuzzyset_selection(
                    universe[feature],
                    membership_functions[feature]["small"],
                    membership_functions[feature]["moderate"],
                    membership_functions[feature]["large"],
                    df_instance[feature],
                )

            fuzzy_sets[df.columns[-1]] = df.loc[i, df.columns[-1]]
            fuzzy_rules.append(fuzzy_sets)

        # Create dataframe of fuzzy rules
        fuzzy_rules_df = pd.DataFrame(fuzzy_rules, index=df.index)

        return fuzzy_rules_df

    def _contextual_synergy_analysis(self, fuzzy_rules_df):
        """
        Identify most occuring fuzzy rules by context (e.g. target category:low, medium, high)
        Parameters:
            fuzzy_rules_df (pd.DataFrame): Features with fuzzy rules.
        Returns:
            pd.DataFrame: Most occuring fuzzy rules.
        """
        self._logger.info(
            "Use most occuring fuzzy rules to extract synergy of features..."
        )
        # Drop rules with all NaN values
        fuzzy_rules_df.dropna(how="all", axis=1, inplace=True)

        # Group by all columns to count the number of occurences of each rule
        fuzzy_rules_grouped = (
            fuzzy_rules_df.groupby(fuzzy_rules_df.columns.tolist())
            .size()
            .reset_index()
            .rename(columns={0: "records"})
        )

        # Sort by the number of occurences
        fuzzy_rules_grouped = fuzzy_rules_grouped.sort_values(
            by=["records"], ascending=False
        )

        # Identify top n rules in each target category
        top_rules = {}
        for category in fuzzy_rules_grouped[fuzzy_rules_grouped.columns[-2]].unique():
            top_rules[category] = fuzzy_rules_grouped[
                fuzzy_rules_grouped[fuzzy_rules_grouped.columns[-2]] == category
            ].head(self._fuzzy_opt.number_rules)

        synergy_features = {}
        for category, rules in top_rules.items():
            synergy_features[category] = {}
            for feature in rules.columns[:-2]:
                unique_values = rules[feature].unique()
                if "large" in unique_values:
                    top_value = "large"
                elif "moderate" in unique_values:
                    top_value = "moderate"
                else:
                    top_value = "small"
                synergy_features[category][feature] = top_value

        self._logger.info(f"synergy and impact of features: \n{synergy_features}")
        # Convert dictionary to dataframe
        synergy_features_df = pd.DataFrame(synergy_features)
        return synergy_features_df



from dataclasses import dataclass

"""In this file, we define the PlottingOptions dataclass, which is used to store the plotting
preferences, defined in the UI for the plots generated by the Helix package. All plots
will then have the same style, such as font size, font family, colour scheme, etc."""


@dataclass
class PlottingOptions:
    plot_axis_font_size: int
    plot_axis_tick_size: int
    plot_colour_scheme: str
    dpi: int
    angle_rotate_xaxis_labels: int
    angle_rotate_yaxis_labels: int
    save_plots: bool
    plot_title_font_size: int
    plot_font_family: str
    height: int | float
    width: int | float
    plot_colour_map: str = "viridis"
    plot_title: str | None = None
    yaxis_label: str | None = None
    xaxis_label: str | None = None
    plot_colour: str | None = None



import json
import os
from pathlib import Path
from pickle import UnpicklingError, dump, load
from typing import TypeVar

from pandas import DataFrame
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin

from helix.options.choices.ml_models import CLASSIFIERS, REGRESSORS
from helix.options.enums import ProblemTypes
from helix.utils.utils import create_directory

MlModel = TypeVar("MlModel", BaseEstimator, ClassifierMixin, RegressorMixin)


def save_model_predictions(predictions: DataFrame, path: Path):
    """Save the predictions of the models to the given file path.

    Args:
        predictions (DataFrame): The predictions to save.
        path (Path): The file path to save the predictions.
    """
    create_directory(path.parent)

    predictions.to_csv(path, index=False)


def save_models_metrics(metrics: dict, path: Path):
    """Save the statistical metrics of the models to the given file path.

    Args:
        metrics (dict): The metrics to save.
        path (Path): The file path to save the metrics.
    """

    create_directory(path.parent)
    with open(path, "w") as f:
        json.dump(metrics, f, indent=4)


def save_model(model, path: Path):
    """Save a machine learning model to the given file path.

    Args:
        model (_type_): The model to save. Must be picklable.
        path (Path): The file path to save the model.
    """
    path.parent.mkdir(exist_ok=True, parents=True)
    with open(path, "wb") as f:
        dump(model, f, protocol=5)


def load_models(path: Path) -> dict[str, list]:
    """Load pre-trained machine learning models.

    Args:
        path (Path): The path to the directory where the models are saved.

    Returns:
        dict[str, list]: The pre-trained models.
    """
    models: dict[str, list] = dict()
    for file_name in sorted(path.iterdir()):
        try:
            with open(file_name, "rb") as file:
                model = load(file)
                model_name = str(file_name).split("/")[-1]
                models[model_name] = model
        except UnpicklingError:
            pass  # ignore bad files

    return models


def load_models_to_explain(path: Path, model_names: list) -> dict[str, list]:
    """Load pre-trained machine learning models.

    Args:
        path (Path): The path to the directory where the models are saved.
        model_names (str): The name of the models to explain.

    Returns:
        dict[str, list]: The pre-trained models.
    """
    models: dict[str, list] = dict()
    for file_name in path.iterdir():
        if os.path.basename(file_name) in model_names or model_names == "all":
            try:
                with open(file_name, "rb") as file:
                    model = load(file)
                    model_name = model.__class__.__name__
                    if model_name in models:
                        models[model_name].append(model)
                    else:
                        models[model_name] = [model]
            except UnpicklingError:
                pass  # ignore bad files
    return models


def get_model_type(model_type: str, problem_type: ProblemTypes) -> type:
    """
    Fetch the appropriate type for a given model name based on the problem type.

    Args:
        model_type (dict): The kind of model.
        problem_type (ProblemTypes): Type of problem (classification or regression).

    Raises:
        ValueError: If a model type is not recognised or unsupported.

    Returns:
        type: The constructor for a machine learning model class.
    """
    if problem_type.lower() == ProblemTypes.Classification:
        model_class = CLASSIFIERS.get(model_type.lower())
    elif problem_type.lower() == ProblemTypes.Regression:
        model_class = REGRESSORS.get(model_type.lower())
    if not model_class:
        raise ValueError(f"Model type {model_type} not recognised")

    return model_class


def models_exist(path: Path) -> bool:
    try:
        trained_models = load_models(path)

        if trained_models:
            return True
        else:
            return False

    except Exception:
        return False


def get_model(model_type: type, model_params: dict = None) -> MlModel:
    """Get a new instance of the requested machine learning model.

    If the model is to be used in a grid search, specify `model_params=None`.

    Args:
        model_type (type): The Python type (constructor) of the model to instantiate.
        model_params (dict, optional): The parameters to pass to the model constructor. Defaults to None.

    Returns:
        MlModel: A new instance of the requested machine learning model.
    """

    return model_type(**model_params) if model_params is not None else model_type()



import torch
import torch.nn.init as init


def normal_init(m: torch.nn.Module, mean: float = 0.0, std: float = 0.02) -> None:
    """
    Initializes the weights of Linear layers using a normal distribution.

    Args:
        m (torch.nn.Module): The module to initialize.
        mean (float): The mean of the normal distribution. Defaults to 0.0.
        std (float): The standard deviation of the normal distribution.
        Defaults to 0.02.

    Returns:
        None
    """
    if isinstance(m, torch.nn.Linear):
        init.normal_(m.weight, mean=mean, std=std)
        if m.bias is not None:
            init.constant_(m.bias, 0)


def xavier_init(m: torch.nn.Module) -> None:
    """
    Initializes the weights of Linear layers using
    Xavier initialization.

    Args:
        m (torch.nn.Module): The module to initialize.

    Returns:
        None
    """
    if isinstance(m, torch.nn.Linear):
        init.xavier_normal_(m.weight)
        if m.bias is not None:
            init.constant_(m.bias, 0)


def kaiming_init(m: torch.nn.Module, nonlinearity: str = "relu") -> None:
    """
    Initializes the weights of Linear layers using Kaiming initialization.

    Args:
        m (torch.nn.Module): The module to initialize.
        nonlinearity (str): The nonlinearity used in the network
        (e.g., 'relu', 'leaky_relu'). Defaults to "relu".

    Returns:
        None
    """
    if isinstance(m, torch.nn.Linear):
        init.kaiming_normal_(m.weight, nonlinearity=nonlinearity)
        if m.bias is not None:
            init.constant_(m.bias, 0)



import logging
import time
from pathlib import Path


class Logger(object):
    """
    logger preparation


    Parameters
    ----------
    log_dir: Path
        path to the log directory

    logging_level: str
        required Level of logging. INFO, WARNING or ERROR can be selected.
        Default to 'INFO'

    console_logger: bool
        flag if console_logger is required. Default to False

    Returns
    ----------
    logger: logging.Logger
        logger object
    """

    def __init__(
        self,
        log_dir: Path | None = None,
        logging_level: str = "INFO",
        console_logger: bool = True,
        multi_module: bool = True,
    ) -> None:
        super().__init__()
        self._log_dir = log_dir
        self.console_logger = console_logger
        self.logging_level = logging_level.lower()
        self.multi_module = multi_module
        self._make_level()

    def _make_level(self):
        if self.logging_level == "info":
            self._level = logging.INFO
        elif self.logging_level == "warning":
            self._level = logging.WARNING
        elif self.logging_level == "error":
            self._level = logging.ERROR
        else:
            raise ValueError(
                "logging_level not specified correctly. "
                "INFO, WARNING or ERROR must be chosen"
            )

    def make_logger(self):
        # Create a custom logger
        if self.multi_module:
            logger = logging.getLogger()
        else:
            logger = logging.getLogger(__name__)
        logger.setLevel(self._level)

        # Create formatter
        format = logging.Formatter("%(levelname)s - %(message)s - %(module)s")

        # Create file handler
        if self._log_dir is not None:
            # logging configuration
            self._log_dir.mkdir(parents=True, exist_ok=True)
            file_name = self._log_dir.joinpath(f"{time.strftime('%Y%m%d-%H%M%S')}.log")
            f_handler = logging.FileHandler(filename=file_name)
            f_handler.setLevel(self._level)
            f_handler.setFormatter(format)
            logger.addHandler(f_handler)

        # Console handler creation
        if self.console_logger:
            c_handler = logging.StreamHandler()
            c_handler.setLevel(self._level)
            c_handler.setFormatter(format)
            logger.addHandler(c_handler)

        return logger

    def remove_handlers(self, logger):
        """
        Remove all handlers associated with the logger object.
        """
        for handler in list(logger.handlers):
            logger.removeHandler(handler)
            handler.close()


def close_logger(logger_instance, logger):
    logger_instance.remove_handlers(logger)
    del logger_instance
    del logger
    return None



import pandas as pd
import streamlit as st


@st.experimental_fragment
def preprocessed_view(data: pd.DataFrame):
    """Display the preprocessed data to the user with column count in titles.

    Args:
        data (pd.DataFrame): The preprocessed data to show.
    """
    st.write("### Processed data")
    st.write(
        f"#### {len(data.columns)-1} independent variables and {len(data)} instances"
    )
    st.write(data)
    st.write("### Processed data description")
    st.write(data.describe())


@st.experimental_fragment
def original_view(data: pd.DataFrame):
    """Display the original data to the user.

    Args:
        data (pd.DataFrame): The original data to show.
    """
    st.write("### Original data")
    st.write(
        f"#### {len(data.columns)-1} independent variables and {len(data)} instances"
    )
    st.write(data)
    st.write("### Original data description")
    st.write(data.describe())



from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
from sklearn.feature_selection import VarianceThreshold
from sklearn.linear_model import Lasso
from sklearn.preprocessing import LabelEncoder, MinMaxScaler, StandardScaler

from helix.options.enums import (
    DataPreprocessingStateKeys,
    Normalisations,
    TransformationsY,
)
from helix.options.file_paths import data_preprocessing_options_path
from helix.options.preprocessing import PreprocessingOptions
from helix.preprocessing.variable_scaling import MeanCenterPoissonScaler, ParetoScaler
from helix.services.configuration import save_options


def find_non_numeric_columns(data: pd.DataFrame | pd.Series) -> List[str]:
    """
    Find non-numeric columns in a DataFrame or check if a Series contains non-numeric values.

    Args:
        data (Union[pd.DataFrame, pd.Series]): The DataFrame or Series to check.

    Returns:
        List[str]: If `data` is a DataFrame, returns a list of non-numeric column names.
                   If `data` is a Series, returns ["Series"] if it contains non-numeric values, else an empty list.
    """
    if isinstance(data, pd.Series):  # If input is a Series
        return (
            data.name
            if data.apply(lambda x: pd.to_numeric(x, errors="coerce")).isna().any()
            else []
        )

    elif isinstance(data, pd.DataFrame):  # If input is a DataFrame
        return [
            col
            for col in data.columns
            if data[col].apply(lambda x: pd.to_numeric(x, errors="coerce")).isna().any()
        ]

    else:
        raise TypeError("Input must be a pandas DataFrame or Series.")


def normalise_independent_variables(normalisation_method: str, X):
    """
    Normalise the independent variables based on the selected method.

    Args:
        normalisation_method (str): The normalisation method to use.
        X (pd.DataFrame): The independent variables to normalise.

    Returns:
        pd.DataFrame: The normalised independent variables.
    """

    match normalisation_method:
        case Normalisations.NoNormalisation:
            return X

        case Normalisations.Standardisation:
            scaler = StandardScaler()

        case Normalisations.MinMax:
            scaler = MinMaxScaler()

        case Normalisations.MeanCentering:
            scaler = StandardScaler(with_std=False)

        case Normalisations.MeanCenteringPoissonScaling:
            scaler = MeanCenterPoissonScaler()

        case Normalisations.ParetoScaling:
            scaler = ParetoScaler()

        case _:
            raise ValueError(
                f"Unsupported normalisation method: {normalisation_method}"
            )

    column_names = X.columns
    processed_X = scaler.fit_transform(X)

    processed_X = pd.DataFrame(processed_X, columns=column_names)

    return processed_X


def transform_dependent_variable(transformation_y_method: str, y):
    """
    Transform the dependent variable based on the selected method.

    Args:
        transformation_y_method (str): The transformation method to use.
        y (pd.Series): The dependent variable to transform.

    Returns:
        pd.Series: The transformed dependent variable.
    """

    if transformation_y_method == TransformationsY.NoTransformation:
        return y

    column_name = y.name
    y = y.to_numpy().reshape(-1, 1)

    if transformation_y_method == TransformationsY.Log:
        if y.min() <= 0:
            y = y - y.min() + 1
        y = np.log(y)

    elif transformation_y_method == TransformationsY.Sqrt:
        if y.min() < 0:
            y = y - y.min()
        y = np.sqrt(y)

    elif transformation_y_method == TransformationsY.MinMaxNormalisation:
        scaler = MinMaxScaler()
        y = scaler.fit_transform(y)

    elif transformation_y_method == TransformationsY.StandardisationNormalisation:
        scaler = StandardScaler()
        y = scaler.fit_transform(y)

    y = pd.DataFrame(y, columns=[column_name])

    return y


def run_feature_selection(
    preprocessing_opts: PreprocessingOptions, data: pd.DataFrame
) -> pd.DataFrame:
    """
    Run feature selection on the data based on the selected methods.

    Args:
        feature_selection_methods (dict): A dictionary of the feature selection methods to use.
        data (pd.DataFrame): The data to perform feature selection on.

    Returns:
        pd.DataFrame: The processed data.

    """
    feature_selection_methods = preprocessing_opts.feature_selection_methods
    X = data.iloc[:, :-1]
    y = data.iloc[:, -1]

    if feature_selection_methods[DataPreprocessingStateKeys.VarianceThreshold]:
        varianceselector = VarianceThreshold(preprocessing_opts.variance_threshold)
        X = varianceselector.fit_transform(X)
        variance_columns = varianceselector.get_feature_names_out()
        X = pd.DataFrame(X, columns=variance_columns)

    if feature_selection_methods[DataPreprocessingStateKeys.CorrelationThreshold]:
        corr_matrix = X.corr().abs()
        upper_triangle = corr_matrix.where(
            np.triu(np.ones(corr_matrix.shape), k=1).astype(bool)
        )
        to_drop = [
            column
            for column in upper_triangle.columns
            if any(upper_triangle[column] > preprocessing_opts.correlation_threshold)
        ]
        X = X.drop(columns=to_drop)

    if feature_selection_methods[DataPreprocessingStateKeys.LassoFeatureSelection]:
        lasso = Lasso(alpha=preprocessing_opts.lasso_regularisation_term)
        lasso.fit(X, y)
        selected_features = X.columns[lasso.coef_ != 0]
        if selected_features.empty:
            raise ValueError(
                "No indepdendent variables remain after applying "
                "LASSO regularisation term "
                f"**{preprocessing_opts.lasso_regularisation_term}**."
            )
        X = X[selected_features]

    processed_data = pd.concat([X, y], axis=1)

    return processed_data


def run_preprocessing(
    data: pd.DataFrame,
    experiment_path: Path,
    config: PreprocessingOptions,
    id_col: str | None = None,
) -> pd.DataFrame:

    X = data.iloc[:, :-1]
    y = data.iloc[:, -1]

    # Save the ID column to add back at the end.
    # Because of how the data are processed on upload, the IDs column
    # should be at index 0
    if id_col is not None:
        ids = X.iloc[:, 0].copy()
    else:
        ids = None

    try:
        columns_to_drop = [id_col] if id_col is not None else []
        columns_to_drop.extend(find_non_numeric_columns(X))
        if columns_to_drop:
            X = X.drop(columns=columns_to_drop)
    except TypeError as e:
        raise e

    try:
        convert_y = find_non_numeric_columns(y)
        if convert_y:
            le = LabelEncoder()
            y = le.fit_transform(y)
            y = pd.Series(y, name=convert_y)
    except TypeError as e:
        raise e

    save_options(data_preprocessing_options_path(experiment_path), config)
    X = normalise_independent_variables(config.independent_variable_normalisation, X)
    y = transform_dependent_variable(config.dependent_variable_transformation, y)

    normalised_data = pd.concat([X, y], axis=1)

    processed_data = run_feature_selection(config, normalised_data)

    # If there was an ID column, put it back
    if id_col is not None:
        processed_data = pd.concat([ids, processed_data], axis=1)

    return processed_data


# Function to convert nominal columns to numeric
def convert_nominal_to_numeric(data: pd.DataFrame) -> pd.DataFrame:
    """Convert all nominal (categorical) columns in a DataFrame to numeric values.
    This function identifies all object or category type columns in the input DataFrame
    and converts them to numeric representations using pandas' factorize method.
    Each unique category is assigned a unique integer value.

    Args:
        data (pd.DataFrame): The input DataFrame containing columns to be converted.

    Returns:
        pd.DataFrame: A DataFrame with all categorical columns converted to numeric values.
    """
    for col in data.columns:
        if (
            data[col].dtype == "object" or data[col].dtype.name == "category"
        ):  # Check if column is categorical
            data[col] = pd.factorize(data[col])[0]
    return data



from enum import StrEnum


class MachineLearningStateKeys(StrEnum):
    """Enum for the state keys related to machine learning."""

    ModelTypes = "model_types"
    SaveModels = "save_models"
    RerunML = "rerun_ml"
    MLLogBox = "ml_log_box"
    Predictions = "predictions"


class ModelFormStateKeys(StrEnum):
    """ "Enum for the state keys in the ML forms."""

    InterceptLinearRegression = "intercept_linear_regression"
    InterceptLasso = "intercept_lasso"
    AlphaLasso = "alpha_lasso"
    InterceptRidge = "intercept_ridge"
    AlphaRidge = "alpha_ridge"
    InterceptElasticNet = "intercept_elastic_net"
    AlphaElasticNet = "alpha_elastic_net"


class FeatureImportanceStateKeys(StrEnum):
    """Enum for the state keys related to feature importance."""

    NumberOfImportantFeatures = "num_important_features"
    ScoringFunction = "scoring_function"
    NumberOfRepetitions = "num_repetitions"
    SaveFeatureImportanceOptions = "save_feature_importance_options"
    SaveFeatureImportanceResults = "save_feature_importance_results"
    LocalImportanceFeatures = "local_importance_methods"
    EnsembleMethods = "ensemble_methods"
    GlobalFeatureImportanceMethods = "global_feature_importance_methods"
    ExplainModels = "explain_models"
    ExplainAllModels = "explain_all_models"
    FILogBox = "fi_log_box"


class FuzzyStateKeys(StrEnum):
    """Enum for the state keys related to fuzzy importance."""

    FuzzyFeatureSelection = "fuzzy_feature_selection"
    NumberOfFuzzyFeatures = "num_fuzzy_features"
    GranularFeatures = "granular_features"
    NumberOfClusters = "num_clusters"
    ClusterNames = "cluster_names"
    NumberOfTopRules = "num_top_rules"
    RerunFI = "rerun_fi"
    FuzzyLogBox = "fuzzy_log_box"


class ExecutionStateKeys(StrEnum):
    """Enum for state keys related to the execution of experiments."""

    ExperimentName = "experiment_name"
    DependentVariable = "dependent_variable"
    DependentVariableName = "dependent_variable_name"
    FeatureColumns = "feature_columns"
    IdColumn = "id_column"
    UploadedFileName = "uploaded_file_name"
    RandomSeed = "random_seed"
    UseHyperParamSearch = "use_hyperparam_search"
    ProblemType = "problem_type"
    DataSplit = "data_split"
    NumberOfBootstraps = "num_bootstraps"
    Normalisation = "normalisation"
    NCPUs = "n_cpus"


class DataAnalysisStateKeys(StrEnum):
    """Enum for app state keys relating to the Data Visualisation page."""

    TargetVarDistribution = "target_var_distribution"
    ShowKDE = "show_kde"
    NBins = "n_bins"
    SaveTargetVarDistribution = "save_target_var_distribution"
    CorrelationHeatmap = "correlation_heatmap"
    AdvancedCorrOptions = "advanced_correlation_options"
    DescriptorCorrelation = "descriptor_correlation"
    CalculateCorrelationMatrix = "calculate_correlation_matrix"
    SelectAllDescriptorsCorrelation = "select_all_descriptors_correlation"
    SaveHeatmap = "save_heatmap"
    PairPlot = "pair_plot"
    SelectAllDescriptorsPairPlot = "select_all_descriptors_pair_plot"
    DescriptorPairPlot = "descriptor_pair_plot"
    ExcludeCorner = "exclude_corner"
    PairPlotCorrelationType = "pair_plot_correlation_type"
    PairPlotDiagonalType = "pair_plot_diagonal_type"
    SavePairPlot = "save_pair_plot"
    TSNEPlot = "tsne_plot"
    SelectNormTsne = "select_norm_tsne"
    Perplexity = "perplexity"
    SaveTSNEPlot = "save_tsne_plot"


class DataPreprocessingStateKeys(StrEnum):
    """Enum for app state keys relating to the Data Preprocessing page."""

    DependentNormalisation = "dependent_normalisation"
    IndependentNormalisation = "independent_normalisation"
    ProceedTransformation = "proceed_transformation"
    VarianceThreshold = "variance_threshold"
    ThresholdVariance = "threshold_variance"
    CorrelationThreshold = "correlation_threshold"
    ThresholdCorrelation = "threshold_correlation"
    LassoFeatureSelection = "lasso_feature_selection"
    RegularisationTerm = "regularisation_term"


class OptimiserTypes(StrEnum):
    Adam = "adam"
    SGD = "sgd"
    RMSprop = "rmsprop"


class ActivationFunctions(StrEnum):
    Identity = "identity"
    Logistic = "logistic"
    Tanh = "tanh"
    ReLU = "relu"


class ProblemTypes(StrEnum):
    Auto = "auto"
    Classification = "classification"
    Regression = "regression"
    BinaryClassification = "binary_classification"
    MultiClassification = "multi_classification"


class SvmKernels(StrEnum):
    RBF = "rbf"
    Linear = "linear"
    Poly = "poly"
    Sigmoid = "sigmoid"
    Precomputed = "precomputed"


class Normalisations(StrEnum):
    # changing spelling to UK here would make this not backwards-compatible
    Standardisation = "standardisation"
    MinMax = "minmax"
    NoNormalisation = "none"  # field name can't be None
    MeanCentering = "mean centering"
    MeanCenteringPoissonScaling = "mean centering and poisson scaling"
    ParetoScaling = "pareto scaling"


class TransformationsY(StrEnum):
    Log = "log"
    Sqrt = "square-root"
    MinMaxNormalisation = "minmax"
    StandardisationNormalisation = "standardisation"  # to match normalisations
    NoTransformation = "none"


class ModelNames(StrEnum):
    LinearModel = "linear model"
    RandomForest = "random forest"
    XGBoost = "xgboost"
    SVM = "svm"
    BRNN = "bayesian regularised neural network"
    MLREM = "multiple linear regression with expectation maximisation"
    Lasso = "least absolute shrinkage and selection operator"
    ElasticNet = "elastic net"
    Ridge = "ridge"
    KNearestNeighbours = "k nearest neighbours"
    MLP = "multilayer perceptron"
    GNB = "gaussian naive bayes"


class DataSplitMethods(StrEnum):
    Holdout = "holdout"
    KFold = "k-fold"
    NoSplit = "none"  # field name can't be None


class Metrics(StrEnum):
    Accuracy = "accuracy"
    F1Score = "f1_score"
    Precision = "precision_score"
    Recall = "recall_score"
    ROC_AUC = "roc_auc_score"
    R2 = "R2"
    MAE = "MAE"
    RMSE = "RMSE"


class PlotOptionKeys(StrEnum):
    AxisFontSize = "plot_axis_font_size"
    AxisTickSize = "plot_axis_tick_size"
    ColourScheme = "plot_colour_scheme"
    ColourMap = "plot_colour_map"
    RotateXAxisLabels = "angle_rotate_xaxis_labels"
    RotateYAxisLabels = "angle_rotate_yaxis_labels"
    SavePlots = "save_plots"
    TitleFontSize = "plot_title_font_size"
    FontFamily = "plot_font_family"
    DPI = "dpi"
    Height = "plot_height"
    Width = "plot_width"
    PlotTitle = "plot_title"
    YAxisLabel = "yaxis_label"
    XAxisLabel = "xaxis_label"
    PlotColour = "plot_colour"

    ShowCustomPlotCreator = "show_custom_plot_creator"
    SelectedModels = "selected_models"
    SelectedSets = "selected_sets"
    SelectedFolds = "selected_folds"
    ColorBy = "color_by"
    StyleBy = "style_by"
    PointSize = "point_size"
    PointColour = "point_colour"


class ViewExperimentKeys(StrEnum):
    ExperimentName = "view_experiment_name"
    ShowCustomPlotCreator = "show_custom_plot_creator"
    CustomPlotName = "custom_plot_name"
    SavePlotButton = "save_plot_button"


class FeatureImportanceTypes(StrEnum):
    LIME = "LIME"
    MajorityVote = "Majority Vote"
    Mean = "Mean"
    PermutationImportance = "Permutation importance"
    SHAP = "SHAP"


class PlotTypes(StrEnum):
    TargetVariableDistribution = "target_variable_distribution"
    CorrelationHeatmap = "correlation_heatmap"
    PairPlot = "pair_plot"
    TSNEPlot = "tsne_plot"
    ParityPlot = "parity_plot"
    ConfusionMatrix = "confusion_matrix"
    AUROC = "auc_roc"


class PredictStateKeys(StrEnum):
    PredictFile = "predict_file"
    PredictModels = "predict_models"
    PredictButton = "predict_button"



from pathlib import Path

from helix.feature_importance.interpreter import FeatureImportanceEstimator
from helix.options.data import DataOptions
from helix.options.execution import ExecutionOptions
from helix.options.fi import FeatureImportanceOptions
from helix.options.plotting import PlottingOptions
from helix.services.data import TabularData


def run(
    fi_opt: FeatureImportanceOptions,
    exec_opt: ExecutionOptions,
    plot_opt: PlottingOptions,
    data_opt: DataOptions,
    data: TabularData,
    models,
    data_path: Path,
    logger,
):

    # Interpret the model results
    interpreter = FeatureImportanceEstimator(
        fi_opt=fi_opt,
        exec_opt=exec_opt,
        plot_opt=plot_opt,
        data_opt=data_opt,
        logger=logger,
        data_path=data_path,
    )
    gloabl_importance_results, local_importance_results, ensemble_results = (
        interpreter.interpret(models, data)
    )

    return gloabl_importance_results, local_importance_results, ensemble_results



import numpy as np
import numpy.linalg.linalg as LA
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_array, check_X_y


class EMLinearRegression(RegressorMixin, BaseEstimator):
    """Multiple Linear Regression with Expectation-Maximisation.

    This implementation uses the EM algorithm for feature selection and weight optimisation.

    Parameters
    ----------
    alpha : float, default=0.
        Regularisation parameter.
    max_beta : float, default=50
        Maximum beta value to optimise over.
    weight_threshold : float, default=1e-3
        Threshold for feature removal.
    max_iterations : int, default=300
        Maximum number of EM algorithm iterations.
    tolerance : float, default=0.01
        Convergence tolerance for relative change in SSD.
    """

    def __init__(
        self,
        alpha=0.5,
        max_beta=20,
        weight_threshold=1e-3,
        max_iterations=300,
        tolerance=0.01,
    ):
        self.alpha = alpha
        self.max_beta = max_beta
        self.weight_threshold = weight_threshold
        self.max_iterations = max_iterations
        self.tolerance = tolerance

        # Attributes set during fit
        self.best_beta = None
        self.weights_ = None
        self.coefficients_ = None
        self.intercept_ = None
        self.p_values_ = None

    def fit(self, X, y):
        """Fit the EM linear regression model with optimised beta.

        Args:
        X : array-like of shape (n_samples, n_features)
            Training data
        y : array-like of shape (n_samples,)
            Target values
        feature_names : list of str, optional
            Names of features. If None, will use indices.

        Returns:
        self : object
            Fitted model for the best value of beta within the specified range.
        """
        # Scale inputs
        scaler = StandardScaler()
        X = scaler.fit_transform(X)

        # Input validation
        X, y = check_X_y(X, y, y_numeric=True)
        y = y.reshape(-1, 1)
        n_samples, n_features = X.shape

        # Add intercept
        H = np.ones((n_samples, n_features + 1), float)
        H[:, 1:] = X
        HT = H.T
        HTy = HT @ y
        G = HT @ H
        weights = LA.pinv(HT @ H) @ HTy

        best_beta = None
        best_weights = None
        best_ssd = float("inf")

        # Beta optimisation loop
        for beta in np.logspace(-2, np.log10(self.max_beta), 50):
            weights, ssd = self._em_algorithm(H, HTy, G, y, beta, weights)
            if ssd < best_ssd:
                best_ssd = ssd
                best_beta = beta
                best_weights = weights

        self.best_beta = best_beta
        self.weights_ = best_weights
        self.intercept_ = best_weights[0, 0]
        self.coefficients_ = best_weights[1:].flatten()
        return self

    def _em_algorithm(self, H, HTy, G, y, beta, weights):
        """Expectation-Maximisation algorithm for feature selection and weight estimation."""
        n_samples, n_features = H.shape

        ssd2 = 1.0
        change = 1.0
        iteration = 0

        while (
            iteration < self.max_iterations and change > self.tolerance and ssd2 > 1e-15
        ):

            iteration += 1
            ssd1 = ssd2

            # ---- 1. Diagonal weights (vector form, no diag matrix needed) ----
            u = np.abs(weights).flatten()

            # Early stop if weights collapse
            if np.all(u < 1e-12):
                break

            # ---- 2. Regularisation matrix Ic ----
            Ic = np.eye(n_features) * (self.alpha + beta * ssd1**2)

            # ---- 3. Build A = Ic + U G U  using vectorized diagonal mult ----
            # A[i,j] = Ic[i,j] + u[i] * G[i,j] * u[j]
            A = Ic + (u[:, None] * G * u[None, :])

            # ---- 4. b = U * HTy (diagonal mult) ----
            b = u[:, None] * HTy

            # ---- 5. Solve A w = b (avoid pinv) ----
            try:
                w_new = np.linalg.solve(A, b)
            except np.linalg.LinAlgError:
                # fallback to pseudo-inverse if A is nearly singular
                w_new = np.linalg.pinv(A) @ b

            # Multiply final U again: weights = U w_new
            weights = u[:, None] * w_new

            # Check for divergence
            if not np.all(np.isfinite(weights)):
                break

            # ---- 6. Compute error ----
            residuals = H @ weights - y
            ssd2 = np.linalg.norm(residuals) / np.sqrt(n_samples)

            # ---- 7. Convergence ----
            change = abs(ssd2 - ssd1) / max(ssd1, 1e-12)

        return weights, ssd2

    def predict(self, X):
        if self.coefficients_ is None:
            raise ValueError("Model must be fitted before making predictions.")
        X = check_array(X)
        scaler = StandardScaler()
        X = scaler.fit_transform(X)
        n_samples = X.shape[0]

        X = np.hstack((np.ones((n_samples, 1)), X))
        predictions = X @ np.vstack(
            ([self.intercept_], self.coefficients_.reshape(-1, 1))
        )
        return predictions.ravel()

    @property
    def coef_(self):
        """Get the coefficients. This property exists for scikit-learn compatibility."""
        return self.coefficients_

    def score(self, X, y, sample_weight=None):
        r2 = super().score(X, y, sample_weight)
        n_samples = X.shape[0]
        n_features = X.shape[1]
        adjusted_r2 = 1 - (1 - r2) * (n_samples - 1) / (n_samples - n_features - 1)
        self.r2_ = r2
        self.adjusted_r2_ = adjusted_r2
        return r2



"""Machine Learning model options forms for Helix.

This module contains form functions for both selecting and configuring different
machine learning models through Streamlit UI elements. Each function handles the
parameter configuration for a specific model type, supporting both manual parameter
setting and hyperparameter search options.

Functions:
    - ml_options_form: The main form for setting up the machine learning pipeline.
    - _linear_model_opts: Configuration for linear regression/classification
    - _random_forest_opts: Configuration for random forest models
    - _xgboost_opts: Configuration for XGBoost models
    - _svm_opts: Configuration for Support Vector Machines
    - _mlrem_opts: Configuration for Multiple Linear Regression with EM
    - _brnn_opts: Configuration for Bayesian Regularised Neural Networks
"""

from multiprocessing import cpu_count

import streamlit as st

from helix.components.configuration import data_split_options_box
from helix.options.choices.ui import SVM_KERNELS
from helix.options.enums import (
    ActivationFunctions,
    ExecutionStateKeys,
    MachineLearningStateKeys,
    ModelFormStateKeys,
    ModelNames,
    PlotOptionKeys,
    ProblemTypes,
)
from helix.options.search_grids import (
    BRNN_GRID,
    ELASTICNET_GRID,
    GNB_GRID,
    KNN_GRID,
    LASSO_GRID,
    LINEAR_MODEL_GRID,
    MLP_GRID,
    MLREM_GRID,
    RANDOM_FOREST_GRID,
    RIDGE_GRID,
    SVM_GRID,
    XGB_GRID,
)


###########################################################################
# Below are the options for which models will be used
###########################################################################
@st.experimental_fragment
def ml_options_form(problem_type: ProblemTypes = ProblemTypes.Regression):
    """
    The form for setting up the machine learning pipeline.
    """

    use_hyperparam_search = st.toggle(
        "Use hyper-parameter search",
        value=True,
        key=ExecutionStateKeys.UseHyperParamSearch,
    )
    if use_hyperparam_search:
        st.success(
            """
            **✨ Hyper-parameters will be searched automatically**.

            Helix will determine the best hyper-parameters
            and return the model with the best performance.
            """
        )
    else:
        st.info(
            """
            **🛠️ Manually set the hyper-parameters you wish to use for your models.**
            """
        )

    data_split_opts = data_split_options_box(not use_hyperparam_search)
    st.session_state[ExecutionStateKeys.DataSplit] = data_split_opts

    max_cpus = max(cpu_count() - 1, 1)
    st.info(
        f"🔎 The number of CPUs you can use is capped at **{max_cpus}/{cpu_count()}**"
        f" to protect your machine."
    )
    st.number_input(
        "Number of CPUs",
        min_value=1,
        max_value=max_cpus,
        value=max_cpus,
        key=ExecutionStateKeys.NCPUs,
        help="The more CPUs you use, faster the training will complete.",
    )

    st.subheader("Select and configure which models to train")
    model_types = {}
    if problem_type == ProblemTypes.Regression:
        string_toggle = "Linear Regression"
        # Only show MLREM for regression problems
        mlrem_model_type = _mlrem_opts(use_hyperparam_search)
        model_types.update(mlrem_model_type)

    else:
        string_toggle = "Logistic Regression"

    if st.toggle(string_toggle, value=False):
        lm_model_type = _linear_model_opts(use_hyperparam_search, problem_type)
        model_types.update(lm_model_type)

    if problem_type == ProblemTypes.Regression:

        # these models only work for regression problems

        lasso_model_type = _lasso_model_opts(use_hyperparam_search)
        model_types.update(lasso_model_type)

        elasticnet_model_type = _elastic_net_model_opts(use_hyperparam_search)
        model_types.update(elasticnet_model_type)

        ridge_model_type = _ridge_model_opts(use_hyperparam_search)
        model_types.update(ridge_model_type)

    else:

        # GNB only works for classification problems
        gnb_model_type = _gnb_model_opts(use_hyperparam_search)
        model_types.update(gnb_model_type)

    knn_model_type = _knn_model_opts(use_hyperparam_search)
    model_types.update(knn_model_type)

    rf_model_type = _random_forest_opts(use_hyperparam_search, problem_type)
    model_types.update(rf_model_type)

    xgb_model_type = _xgboost_opts(use_hyperparam_search)
    model_types.update(xgb_model_type)

    svm_model_type = _svm_opts(use_hyperparam_search, problem_type)
    model_types.update(svm_model_type)

    mlp_model_type = _mlp_model_opts(use_hyperparam_search)
    model_types.update(mlp_model_type)

    # if st.toggle("Bayesian Regularised Neural Network", value=False):
    #     brnn_model_type = _brnn_opts(use_hyperparam_search)
    #     model_types.update(brnn_model_type)

    st.session_state[MachineLearningStateKeys.ModelTypes] = model_types
    st.subheader("Select outputs to save")
    st.toggle(
        "Save models",
        key=MachineLearningStateKeys.SaveModels,
        value=True,
        help="Save the models that are trained to disk?",
    )
    st.toggle(
        "Save plots",
        key=PlotOptionKeys.SavePlots,
        value=True,
        help="Save the plots to disk?",
    )


###########################################################################
# Below are the individual model options functions
###########################################################################
def _linear_model_opts(use_hyperparam_search: bool, problem_type: ProblemTypes) -> dict:
    model_types = {}
    if not use_hyperparam_search:
        st.write("Options:")
        fit_intercept = st.checkbox(
            "Fit intercept", key=ModelFormStateKeys.InterceptLinearRegression
        )
        params = {
            "fit_intercept": fit_intercept,
        }
        if problem_type == ProblemTypes.Classification:
            params["class_weight"] = "balanced"
        st.divider()
    else:
        params = LINEAR_MODEL_GRID
        if problem_type == ProblemTypes.Classification:
            params["class_weight"] = ["balanced"]
    model_types[ModelNames.LinearModel.value] = {
        "use": True,
        "params": params,
    }
    return model_types


def _lasso_model_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}

    if st.toggle("LASSO", value=False):
        if not use_hyperparam_search:
            st.write("Options:")
            fit_intercept = st.checkbox(
                "Fit intercept", key=ModelFormStateKeys.InterceptLasso
            )
            alpha = st.number_input(
                "Alpha (regularisation term)",
                value=0.05,
                min_value=0.0,
                step=0.05,
                key=ModelFormStateKeys.AlphaLasso,
            )
            params = {"fit_intercept": fit_intercept, "alpha": alpha}
            st.divider()
        else:
            params = LASSO_GRID

        model_types[ModelNames.Lasso.value] = {
            "use": True,
            "params": params,
        }

    return model_types


def _ridge_model_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}

    if st.toggle("Ridge", value=False):
        if not use_hyperparam_search:
            st.write("Options:")
            fit_intercept = st.checkbox(
                "Fit intercept", key=ModelFormStateKeys.InterceptRidge
            )
            alpha = st.number_input(
                "Alpha (regularisation term)",
                value=0.05,
                min_value=0.0,
                step=0.05,
                key=ModelFormStateKeys.AlphaRidge,
            )
            params = {"fit_intercept": fit_intercept, "alpha": alpha}
            st.divider()
        else:
            params = RIDGE_GRID

        model_types[ModelNames.Ridge.value] = {
            "use": True,
            "params": params,
        }

    return model_types


def _elastic_net_model_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}
    if st.toggle("Elastic-Net", value=False):

        if not use_hyperparam_search:
            st.write("Options:")
            fit_intercept = st.checkbox(
                "Fit intercept", key=ModelFormStateKeys.InterceptElasticNet
            )
            alpha = st.number_input(
                "Alpha (regularisation term)",
                value=0.05,
                min_value=0.0,
                step=0.05,
                key=ModelFormStateKeys.AlphaElasticNet,
            )
            l1_ratio = st.slider(
                "L1 ratio",
                min_value=0.0,
                max_value=1.0,
                value=0.5,
                step=0.1,
                help="Select based on how much L1 penalty to use for training. A value of 1 is 100% L1, a value of 0.5 is 50% L1 and 50% L2 and 0 100% L2.",
            )
            params = {
                "fit_intercept": fit_intercept,
                "alpha": alpha,
                "l1_ratio": l1_ratio,
            }
            st.divider()
        else:
            params = ELASTICNET_GRID

        model_types[ModelNames.ElasticNet.value] = {
            "use": True,
            "params": params,
        }

    return model_types


def _knn_model_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}

    if st.toggle("K-Nearest Neighbours", value=False):
        if not use_hyperparam_search:
            st.write("Options:")
            n_neighbors = st.number_input(
                "Number of neighbours",
                value=5,
                min_value=3,
                max_value=15,
                step=1,
                help="Number of neighbors to use by default.",
            )
            leaf_size = st.number_input(
                "Leaf size",
                value=30,
                min_value=5,
                max_value=50,
                step=5,
                help="Leaf size passed to BallTree or KDTree. This can affect the speed of the construction and query, as well as the memory required to store the tree. The optimal value depends on the nature of the problem.",
            )
            p = st.slider(
                "p value",
                min_value=1,
                max_value=2,
                value=1,
                step=1,
                help="Power parameter for the Minkowski metric. When p = 1, this is equivalent to using manhattan_distance (l1), and euclidean_distance (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used. This parameter is expected to be positive.",
            )
            params = {"n_neighbors": n_neighbors, "leaf_size": leaf_size, "p": p}
            st.divider()
        else:
            params = KNN_GRID

        model_types[ModelNames.KNearestNeighbours.value] = {
            "use": True,
            "params": params,
        }

    return model_types


def _gnb_model_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}
    if st.toggle("Gaussian Naive Bayes"):
        if not use_hyperparam_search:
            st.write("Options:")
            var_smoothing = st.selectbox(
                "Variable Smoothing",
                options=[1e-09, 1e-08],
                index=0,
                help="Portion of the largest variance of all features that is added to variances for calculation stability.",
            )
            params = {
                "var_smoothing": var_smoothing,
            }
            st.divider()
        else:
            params = GNB_GRID
        model_types[ModelNames.GNB.value] = {
            "use": True,
            "params": params,
        }

    return model_types


def _mlp_model_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}

    if st.toggle("Multilayer Perceptron", value=False):

        if not use_hyperparam_search:
            st.write("Options:")
            n_neurons = st.number_input(
                "Number of neurons per layer",
                value=75,
                min_value=1,
                max_value=None,
                step=1,
            )
            num_hidden_layers = st.number_input(
                "Number of hidden layers", value=1, min_value=1, max_value=None, step=1
            )

            hidden_layer_sizes = tuple([n_neurons] * num_hidden_layers)

            activation = st.selectbox(
                "Activation function", options=["logistic", "tanh", "relu"], index=0
            )
            solver = st.selectbox("Solver", options=["sgd", "adam"], index=1)
            learning_rate_init = st.number_input(
                "Initial learning rate",
                min_value=0.005,
                max_value=0.1,
                value=0.005,
                step=0.005,
            )
            learning_rate = st.selectbox(
                "Learning rate", options=["constant", "adaptive"], index=1
            )
            max_iter = st.slider(
                "Number of epochs (training iterations)",
                min_value=100,
                max_value=300,
                value=150,
                step=10,
            )
            early_stopping = st.toggle("Allow early stopping", value=False)

            alpha = st.number_input(
                "L2 regularisation term strength ",
                min_value=0.0,
                max_value=None,
                value=1e-4,
                step=1e-4,
                format="%.4f",
            )

            params = {
                "hidden_layer_sizes": hidden_layer_sizes,
                "activation": activation,
                "solver": solver,
                "alpha": alpha,
                "learning_rate_init": learning_rate_init,
                "learning_rate": learning_rate,
                "max_iter": max_iter,
                "early_stopping": early_stopping,
            }
            st.divider()
        else:
            params = MLP_GRID

        model_types[ModelNames.MLP.value] = {
            "use": True,
            "params": params,
        }

    return model_types


def _random_forest_opts(
    use_hyperparam_search: bool, problem_type: ProblemTypes
) -> dict:
    model_types = {}
    if st.toggle("Random Forest", value=False):
        if not use_hyperparam_search:
            st.write("Options:")
            n_estimators_rf = st.number_input(
                "Number of estimators", value=100, key="n_estimators_rf"
            )
            min_samples_split = st.number_input("Minimum samples split", value=2)
            min_samples_leaf = st.number_input("Minimum samples leaf", value=1)
            col1, col2 = st.columns(
                [0.25, 0.75], vertical_alignment="bottom", gap="small"
            )
            use_rf_max_depth = col1.checkbox(
                "Set max depth",
                value=False,
                help="If disabled or 0, then nodes are expanded until all leaves are pure"
                " or until all leaves contain less than 'Minimum samples split'.",
            )
            max_depth_rf = col2.number_input(
                "Maximum depth",
                value="min",
                min_value=0,
                key="max_depth_rf",
                disabled=not use_rf_max_depth,
            )
            params = {
                "n_estimators": n_estimators_rf,
                "min_samples_split": min_samples_split,
                "min_samples_leaf": min_samples_leaf,
                "max_depth": max_depth_rf if max_depth_rf > 0 else None,
            }
            if problem_type == ProblemTypes.Classification:
                params["class_weight"] = "balanced"
            st.divider()
        else:
            params = RANDOM_FOREST_GRID
            if problem_type == ProblemTypes.Classification:
                params["class_weight"] = ["balanced"]

        model_types[ModelNames.RandomForest.value] = {
            "use": True,
            "params": params,
        }

    return model_types


def _xgboost_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}
    if st.toggle("XGBoost", value=False):
        if not use_hyperparam_search:
            if st.checkbox("Set XGBoost options"):
                st.write("Options:")
                n_estimators_xgb = st.number_input(
                    "Number of estimators", value=100, key="n_estimators_xgb"
                )
                learning_rate = st.number_input("Learning rate", value=0.01)
                subsample = st.number_input("Subsample size", value=0.5)
                col1, col2 = st.columns(
                    [0.25, 0.75], vertical_alignment="bottom", gap="small"
                )
                use_xgb_max_depth = col1.checkbox(
                    "Set max depth",
                    value=False,
                    help="If disabled or 0, then nodes are expanded until all leaves are pure.",
                )
                max_depth_xbg = col2.number_input(
                    "Maximum depth",
                    value="min",
                    min_value=0,
                    key="max_depth_xgb",
                    disabled=not use_xgb_max_depth,
                )
            else:
                n_estimators_xgb = None
                max_depth_xbg = None
                learning_rate = None
                subsample = None
            params = {
                "kwargs": {
                    "n_estimators": n_estimators_xgb,
                    "max_depth": max_depth_xbg,
                    "learning_rate": learning_rate,
                    "subsample": subsample,
                }
            }
            st.divider()
        else:
            params = XGB_GRID

        model_types[ModelNames.XGBoost.value] = {
            "use": True,
            "params": params,
        }
    return model_types


def _svm_opts(use_hyperparam_search: bool, problem_type: ProblemTypes) -> dict:
    model_types = {}

    if st.toggle("Support Vector Machine", value=False):
        if not use_hyperparam_search:
            st.write("Options:")
            kernel = st.selectbox("Kernel", options=SVM_KERNELS)
            degree = st.number_input("Degree", min_value=0, value=3)
            c = st.number_input("C", value=1.0, min_value=0.0)
            params = {
                "kernel": kernel.lower(),
                "degree": degree,
                "C": c,
            }
            if problem_type == ProblemTypes.Classification:
                params["class_weight"] = "balanced"
            st.divider()
        else:
            params = SVM_GRID
            if problem_type == ProblemTypes.Classification:
                params["class_weight"] = ["balanced"]

        model_types[ModelNames.SVM.value] = {
            "use": True,
            "params": params,
        }
    return model_types


def _mlrem_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}

    if st.toggle(
        "Multiple Linear Regression with Expectation Maximisation",
        value=False,
        help="MLREM is only available for regression problems. Note that if you run hyperparamenter optimisation, the processing time will take longer",
    ):
        if not use_hyperparam_search:
            st.write("Options")
            # Basic parameters
            alpha = st.number_input(
                "Alpha (regularisation)", value=0.05, min_value=0.05, step=1.0
            )
            max_beta = st.number_input(
                "Maximum Beta",
                value=40,
                help="Will test beta values from 0.1 to max_beta",
            )
            weight_threshold = st.number_input(
                "Weight Threshold",
                value=1e-3,
                min_value=1e-3,
                step=1e-4,
                format="%.4f",
                help="Features with weights below this will be removed",
            )

            # Advanced options
            st.write("Advanced Options:")
            max_iterations = st.number_input(
                "Max Iterations", value=300, min_value=1, step=50
            )
            tolerance = st.number_input(
                "Tolerance", value=0.01, format="%.4f", step=0.001
            )

            params = {
                "alpha": alpha,
                "max_beta": max_beta,
                "weight_threshold": weight_threshold,
                "max_iterations": max_iterations,
                "tolerance": tolerance,
            }
            st.divider()
        else:
            params = MLREM_GRID

        model_types[ModelNames.MLREM.value] = {
            "use": True,
            "params": params,
        }

    return model_types


def _brnn_opts(use_hyperparam_search: bool) -> dict:
    model_types = {}
    if not use_hyperparam_search:
        st.write("Options")
        # Basic parameters
        num_hidden_layers = st.number_input(
            "Number of Hidden Layers", value=2, min_value=1, step=1
        )
        hidden_layer_sizes = st.number_input(
            "Hidden Units", value=25, min_value=1, step=5
        )
        activation = st.selectbox(
            "Activation Function",
            options=[af.value for af in ActivationFunctions],
            index=0,
        )
        batch_size = st.number_input(
            "Batch Size", value=32, min_value=1, step=1, help="Batch size for training"
        )
        learning_rate = st.number_input(
            "Learning Rate", value=0.01, format="%.4f", step=1e-3
        )
        max_iter = st.number_input("Max Iterations", value=200, min_value=1, step=50)
        random_state = st.number_input("Random State", value=42, min_value=0, step=1)

        params = {
            "hidden_layer_sizes": [
                hidden_layer_sizes for _ in range(num_hidden_layers - 2)
            ],
            "activation": activation,
            "batch_size": batch_size,
            "learning_rate_init": learning_rate,
            "max_iter": max_iter,
            "random_state": random_state,
        }
        st.divider()
    else:
        params = BRNN_GRID

    model_types[ModelNames.BRNN.value] = {
        "use": True,
        "params": params,
    }
    return model_types



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class MeanCenterPoissonScaler(BaseEstimator, TransformerMixin):
    """
    Mean Centering + Poisson Scaling transformer.

    - Mean centering: subtracts feature-wise mean.
    - Poisson scaling: divides by sqrt(mean) of each feature.

    Compatible with scikit-learn pipelines.
    """

    def __init__(self):
        self.mean_ = None
        self.scale_ = None

    def fit(self, X, y=None):
        """Compute mean and Poisson scaling factor for each feature."""
        X = self._validate_data(X, dtype=np.float64, force_all_finite="allow-nan")
        self.mean_ = np.nanmean(X, axis=0)
        # Avoid division by zero in case mean is 0
        self.scale_ = np.sqrt(np.where(self.mean_ == 0, 1, self.mean_))
        return self

    def transform(self, X):
        """Apply mean centering and Poisson scaling."""
        X = self._validate_data(
            X, dtype=np.float64, reset=False, force_all_finite="allow-nan"
        )
        return (X - self.mean_) / self.scale_

    def inverse_transform(self, X):
        """Reconstruct the original data."""
        X = np.asarray(X)
        return X * self.scale_ + self.mean_


class ParetoScaler(BaseEstimator, TransformerMixin):
    """
    Pareto Scaling transformer.

    - Scales each feature by sqrt(std).
    - Unlike standardization, does NOT mean-center the data.

    Compatible with scikit-learn pipelines.
    """

    def __init__(self):
        self.scale_ = None

    def fit(self, X, y=None):
        """Compute Pareto scaling factor for each feature."""
        X = self._validate_data(X, dtype=np.float64, force_all_finite="allow-nan")
        std = np.nanstd(X, axis=0, ddof=0)
        # Avoid division by zero in case std is 0
        self.scale_ = np.sqrt(np.where(std == 0, 1, std))
        return self

    def transform(self, X):
        """Apply Pareto scaling."""
        X = self._validate_data(
            X, dtype=np.float64, reset=False, force_all_finite="allow-nan"
        )
        return X / self.scale_

    def inverse_transform(self, X):
        """Reconstruct the original data."""
        X = np.asarray(X)
        return X * self.scale_



from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
from scipy import stats


def kolmogorov_smirnov_test(
    data: np.ndarray | list, reference_dist: str = "norm"
) -> Tuple[float, float]:
    """
    Perform Kolmogorov-Smirnov test to determine if a sample comes from a reference distribution.
    By default, tests against a normal distribution.

    Args:
        data: Input array of observations to test.
              Can be a numpy array or a list.
        reference_dist: String specifying the reference distribution.
                       Default is 'norm' for normal distribution.
                       Other options include: 'uniform', 'expon', etc.

    Returns:
        Tuple containing:
            - statistic: The test statistic
            - p_value: The p-value for the hypothesis test

    Note:
        - Null hypothesis: the data comes from the specified distribution
        - If p-value < alpha (typically 0.05), reject the null hypothesis
          (data does not come from the specified distribution)
        - If p-value >= alpha, fail to reject the null hypothesis
          (data may come from the specified distribution)
    """
    if isinstance(data, list):
        data = np.array(data)

    if len(data.shape) > 1:
        data = data.flatten()

    # Fit the reference distribution to the data
    if reference_dist == "norm":
        # For normal distribution, we need mean and std
        params = stats.norm.fit(data)
        return stats.kstest(data, reference_dist, args=params)
    else:
        # For other distributions, let scipy handle the parameter fitting
        return stats.kstest(data, reference_dist)


def shapiro_wilk_test(data: Union[np.ndarray, list]) -> Tuple[float, float]:
    """
    Perform Shapiro-Wilk test for normality on the input data.

    The Shapiro-Wilk test tests the null hypothesis that the data was drawn from a
    normal distribution.

    Args:
        data: Input array of observations to test for normality.
              Can be a numpy array or a list.

    Returns:
        Tuple containing:
            - statistic: The test statistic
            - p_value: The p-value for the hypothesis test

    Note:
        - Null hypothesis: the data is normally distributed
        - If p-value < alpha (typically 0.05), reject the null hypothesis
          (data is not normally distributed)
        - If p-value >= alpha, fail to reject the null hypothesis
          (data may be normally distributed)
    """
    if isinstance(data, list):
        data = np.array(data)

    if len(data.shape) > 1:
        data = data.flatten()

    return stats.shapiro(data)


def create_normality_test_table(data: pd.DataFrame) -> Optional[pd.DataFrame]:
    """Create a dataframe with normality test results for numerical columns.

    Args:
        data: Input DataFrame containing the data to test

    Returns:
        DataFrame containing normality test results for each numerical column,
        or None if no valid columns are found
    """
    test_results = []
    numerical_cols = data.select_dtypes(include=[np.number]).columns

    for col in numerical_cols:
        # Skip if all values are the same (no variance)
        if len(data[col].unique()) <= 1:
            continue

        # Perform Shapiro-Wilk test
        sw_stat, sw_p = shapiro_wilk_test(data[col].dropna())

        # Perform Kolmogorov-Smirnov test
        ks_stat, ks_p = kolmogorov_smirnov_test(data[col].dropna())

        test_results.append(
            {
                "Variable": col,
                "Shapiro-Wilk Statistic": round(sw_stat, 3),
                "Shapiro-Wilk p-value": round(sw_p, 3),
                "Kolmogorov-Smirnov Statistic": round(ks_stat, 3),
                "Kolmogorov-Smirnov p-value": round(ks_p, 3),
            }
        )

    return pd.DataFrame(test_results) if test_results else None



import argparse

import pandas as pd

from helix.utils.logging_utils import Logger


def calculate_ensemble_mean(
    feature_importance_dict: dict[str, pd.DataFrame], logger: Logger
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Calculate mean of feature importance results across models and methods.

    Args:
        feature_importance_dict: Dictionary where:
            - Keys are model names
            - Values are DataFrames where:
                - Rows are features
                - Columns are importance methods (SHAP, Permutation, etc.)
                - Values are already normalized between 0 and 1
        logger: Logger instance

    Returns:
        pd.DataFrame: DataFrame with unique features as index and a single column
            containing the overall mean importance across all models, methods, and repetitions.
    """
    logger.info("Calculating Mean ensemble importance...")

    # First calculate mean across methods for each model
    model_means = {}
    for model_name, importance_df in feature_importance_dict.items():
        # Calculate mean across all methods for this model
        model_means[model_name] = importance_df.mean(axis=1)

    # Combine all model means into a DataFrame
    all_model_means = pd.DataFrame(model_means)

    # Calculate mean across all models
    combined_mean = all_model_means.mean(axis=1).to_frame(name="Mean Importance")

    # Group by index (feature names) and calculate mean for repeated features
    grouped = combined_mean.groupby(level=0)
    ensemble_mean = grouped.mean()
    ensemble_mean_std = grouped.std()

    # Sort by importance in descending order
    ensemble_mean = ensemble_mean.sort_values("Mean Importance", ascending=False)
    ensemble_mean_std = ensemble_mean_std.reindex(ensemble_mean.index)

    logger.info("Mean ensemble importance calculated...")
    return ensemble_mean, ensemble_mean_std


def calculate_ensemble_majorityvote(
    feature_importance_dict: dict[str, pd.DataFrame], logger: Logger
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Calculate majority vote of feature importance results.
    For majority vote, each vector in the feature importance matrix has their features ranked based on their importance.
    Subsequently, the final feature importance is the average of the most common rank order for each feature.
    For example, feature Xi has a final rank vector of [1, 1, 1, 2], where each rank rk is established by a different feature importance method k.
    The final feature importance value for feature Xi is the average value from the three feature importance methods that ranked it as 1.

    Args:
        feature_importance_dict: Dictionary where:
            - Keys are model names
            - Values are DataFrames where:
                - Rows are features
                - Columns are importance methods (SHAP, Permutation, etc.)
                - Values are already normalized between 0 and 1
        logger: Logger instance

    Returns:
        pd.DataFrame: DataFrame with features as index and a single column
            containing the majority vote importance.
    """
    logger.info("Calculating Majority Vote ensemble importance...")

    # First combine all model DataFrames into a single DataFrame
    ensemble_majorityvote = pd.concat(feature_importance_dict.values())

    # Return early if no feature importances
    if ensemble_majorityvote.empty:
        return pd.DataFrame(columns=["Majority Vote Importance"])

    # Rank the features based on their importance in each model and method
    rank_feature_importance = ensemble_majorityvote.rank(
        axis=0, method="dense", ascending=False
    )

    # Find the mode (most common) rank for each feature
    mode_rank_feature_importance = rank_feature_importance.mode(axis=1)

    # Find the most common rank order for each feature
    majority_votes_1 = mode_rank_feature_importance[0]

    # Create mask for features at their most common rank
    majority_votes_rank_1 = rank_feature_importance.eq(majority_votes_1, axis=0)
    final_majority_votes = majority_votes_rank_1

    # If there's a second most common rank, include it
    if len(mode_rank_feature_importance.columns) > 1:
        majority_votes_2 = mode_rank_feature_importance[1]
        majority_votes_rank_2 = rank_feature_importance.eq(majority_votes_2, axis=0)
        final_majority_votes = final_majority_votes | majority_votes_rank_2

    # Calculate mean importance where features have their most common rank
    ensemble_majorityvote = (
        ensemble_majorityvote.where(final_majority_votes, 0)
        .mean(axis=1)
        .to_frame(name="Majority Vote Importance")
    )

    # Group by index to handle repeated features
    grouped = ensemble_majorityvote.groupby(level=0)
    ensemble_majorityvote = grouped.mean()
    ensemble_majorityvote_std = grouped.std()

    # Sort by importance in descending order
    ensemble_majorityvote = ensemble_majorityvote.sort_values(
        "Majority Vote Importance", ascending=False
    )
    ensemble_majorityvote_std = ensemble_majorityvote_std.reindex(
        ensemble_majorityvote.index
    )

    logger.info("Majority Vote ensemble importance calculated...")
    return ensemble_majorityvote, ensemble_majorityvote_std


def calculate_ensemble_fuzzy(feature_importance_results, opt: argparse.Namespace):
    raise NotImplementedError("Fuzzy ensemble method is not implemented yet")



from pathlib import Path

import streamlit as st

from helix.components.configuration import display_options
from helix.components.experiments import experiment_selector
from helix.components.forms.forms_plots import (
    correlation_heatmap_form,
    pairplot_form,
    target_variable_dist_form,
    tSNE_plot_form,
)
from helix.components.images.logos import sidebar_logo
from helix.components.statistical_tests import normality_test_view
from helix.options.enums import ExecutionStateKeys
from helix.options.file_paths import (
    data_analysis_plots_dir,
    data_options_path,
    execution_options_path,
    helix_experiments_base_dir,
    plot_options_path,
    preprocessed_data_path,
)
from helix.services.configuration import (
    load_data_options,
    load_execution_options,
    load_plot_options,
)
from helix.services.data import read_data
from helix.services.experiments import get_experiments
from helix.services.preprocessing import convert_nominal_to_numeric
from helix.utils.logging_utils import Logger, close_logger
from helix.utils.utils import create_directory

st.set_page_config(
    page_title="Data Visualisation",
    page_icon=sidebar_logo(),
)

sidebar_logo()

st.header("Data Visualisation")
st.write(
    """
    Here you can visualise your data. This is useful for understanding the distribution of your data, normality tests,
    as well as the correlation between different variables.
    """
)

choices = get_experiments()
experiment_name = experiment_selector(choices)
biofefi_base_dir = helix_experiments_base_dir()


def load_dataset(path_to_raw_data: Path, path_to_preproc_data: Path, logger) -> tuple:
    """Load raw and preprocessed data if available.

    Args:
            path_to_raw_data: Path to raw data file
            path_to_preproc_data: Path to preprocessed data file
        logger: Logger instance

    Returns:
        tuple: (raw_data, preprocessed_data, data_for_tsne)
    """
    raw_data = None
    preprocessed_data = None
    data_tsne = None

    if path_to_raw_data.exists():
        raw_data = read_data(path_to_raw_data, logger)
        raw_data = convert_nominal_to_numeric(raw_data)
        data_tsne = raw_data

    if path_to_preproc_data.exists():
        preprocessed_data = read_data(path_to_preproc_data, logger)
        data_tsne = preprocessed_data

    return raw_data, preprocessed_data, data_tsne


def visualisation_view(data, data_tsne, prefix: str | None = None):
    """Display visualisation of data."""

    if data is not None:
        st.write("### Graphical description")
        st.write("#### Target variable distribution")
        target_variable_dist_form(
            data,
            exec_opt.dependent_variable,
            data_analysis_plot_dir,
            plot_opt,
            key_prefix=prefix,
        )

        st.write("#### Correlation heatmap")
        correlation_heatmap_form(
            data, data_analysis_plot_dir, plot_opt, key_prefix=prefix
        )

        st.write("#### Pairplot")
        pairplot_form(data, data_analysis_plot_dir, plot_opt, key_prefix=prefix)

        st.write("#### t-SNE plot")
        tSNE_plot_form(
            data_tsne,
            exec_opt.random_state,
            data_analysis_plot_dir,
            plot_opt,
            data_opts.normalisation,
            key_prefix=prefix,
        )


if experiment_name:
    logger_instance = Logger()
    logger = logger_instance.make_logger()

    st.session_state[ExecutionStateKeys.ExperimentName] = experiment_name

    # Create a container for configuration options
    config_container = st.container()
    with config_container:
        display_options(biofefi_base_dir / experiment_name)

    path_to_exec_opts = execution_options_path(biofefi_base_dir / experiment_name)

    path_to_plot_opts = plot_options_path(biofefi_base_dir / experiment_name)

    data_analysis_plot_dir = data_analysis_plots_dir(biofefi_base_dir / experiment_name)

    path_to_data_opts = data_options_path(biofefi_base_dir / experiment_name)

    create_directory(data_analysis_plot_dir)

    exec_opt = load_execution_options(path_to_exec_opts)
    plot_opt = load_plot_options(path_to_plot_opts)
    data_opts = load_data_options(path_to_data_opts)

    try:

        # `raw_data` refers to the data before it gets any preprocessing,
        # such as Standardisation or Log transformation.
        # `data` can be preprocessed data, if the user has used the preprocessing page.
        # If not, then `data` will be the raw data that they uploaded.
        # In this case `raw_data` will be None.
        path_to_raw_data = Path(data_opts.data_path.replace("_preprocessed", ""))
        path_to_preproc_data = preprocessed_data_path(
            str(path_to_raw_data), biofefi_base_dir / experiment_name
        )

        # Load data based on what's available
        raw_data, preprocessed_data, data_tsne = load_dataset(
            path_to_raw_data, path_to_preproc_data, logger
        )

        st.write("### Dataset overview")

        # Create tabs based on available data
        if preprocessed_data is not None:
            raw_tab, preprocessed_tab = st.tabs(["Raw data", "Preprocessed data"])
            # I need to use tabs in the interface to show both data
            with raw_tab:
                st.write(
                    f"#### Raw data [{len(raw_data.columns)-1} independent variables]"
                )
                st.info("This is your original data **before** preprocessing.")
                st.dataframe(raw_data)

                st.write("#### Data statistics")
                st.write(raw_data.describe())

                normality_test_view(raw_data, "Raw data")
                # Data visualisation
                visualisation_view(raw_data, data_tsne, prefix="raw")

            # I need to use tabs in the interface to show both data
            with preprocessed_tab:
                st.write(
                    f"#### Preprocessed data [{len(preprocessed_data.columns)-1} independent variables]"
                )
                st.info("This is your original data **after** preprocessing.")
                st.dataframe(preprocessed_data)

                st.write("#### Data statistics")
                st.write(preprocessed_data.describe())

                normality_test_view(preprocessed_data, "Preprocessed data")
                # Data visualisation
                visualisation_view(preprocessed_data, data_tsne, prefix="preprocessed")

        else:  # raw data only available, so no need for tabs
            if raw_data is not None:
                st.write(
                    f"#### Raw data [{len(raw_data.columns)-1} independent variables]"
                )
                st.info("This is your original data **before** preprocessing.")
                st.dataframe(raw_data)

                if preprocessed_data is None:
                    st.write("#### Data statistics")
                    st.write(raw_data.describe())

                normality_test_view(raw_data, "Raw data")
                visualisation_view(raw_data, data_tsne, prefix="raw")
            else:
                st.info("No raw data available.")

    except ValueError:
        # When the user uploaded the wrong file type, somehow
        st.error("You must upload a .csv or .xlsx file.", icon="🔥")
    except Exception:
        # Catch all error
        st.error("Something went wrong.", icon="🔥")

    finally:
        close_logger(logger_instance, logger)



"""Component for displaying statistical test results."""

import pandas as pd
import streamlit as st

from helix.services.statistical_tests import create_normality_test_table


def display_normality_test_results(results_df: pd.DataFrame, title: str):
    """Display normality test results in a formatted table.

    Args:
        results_df (pd.DataFrame): DataFrame containing normality test results
        title (str): Title to display above the results table
    """
    if results_df is not None:
        st.write(f"#### {title}")
        st.write(
            """
        These tests evaluate whether the data follows a normal distribution:
        - If p-value < 0.05: Data is likely not normally distributed
        - If p-value ≥ 0.05: Data might be normally distributed
        """
        )
        st.dataframe(
            results_df.style.format(
                {
                    "Shapiro-Wilk Statistic": "{:.3f}",
                    "Shapiro-Wilk p-value": "{:.3f}",
                    "Kolmogorov-Smirnov Statistic": "{:.3f}",
                    "Kolmogorov-Smirnov p-value": "{:.3f}",
                }
            )
        )


@st.experimental_fragment
def normality_test_view(
    data: pd.DataFrame,
    table_title: str | None = None,
):

    # Get normality test results for raw data
    if data is not None:
        results = create_normality_test_table(data)
        display_normality_test_results(results, f"Normality tests for {table_title}")
    else:
        st.info("No data available.")
